import json
from dataclasses import dataclass, field
from datetime import datetime
from decimal import Decimal
from enum import Enum
from typing import Generic, NamedTuple, List, Dict, TypeVar, Union

import pandas as pd

from .._typing import DemeterError, TokenInfo

T = TypeVar("T")


class Rule(NamedTuple):
    """
    Rule properties
    """

    agg: str | None
    fillna_method: str | None
    fillna_value: int | None


# @dataclass
# class RowData:
#     """
#     Row properties
#     """
#
#     timestamp: datetime = None
#     row_id: int = None


class MarketTypeEnum(Enum):
    uniswap_v3 = 1
    aave_v3 = 2
    deribit_option = 3
    squeeth = 4


class MarketInfo(NamedTuple):
    """
    MarketInfo properties
    """

    name: str  # uni_market
    type: MarketTypeEnum = MarketTypeEnum.uniswap_v3

    def __str__(self):
        return f"{self.name}({self.type.name})"

    def __repr__(self):
        return f"{self.name}({self.type.name})"


class AssetDescription(NamedTuple):
    name: str
    value: float


class Asset(object):
    """
    Wallet of broker, manage balance of an asset.
    It will prevent excess usage on asset.
    """

    def __init__(self, token: TokenInfo, init_amount=Decimal(0)):
        """
        initialization of Asset
        :param token: token info
        :param init_amount: initialization amount, default 0
        """
        self.token_info = token
        self.name = token.name
        self.decimal = token.decimal
        self.balance = init_amount

    def __str__(self):
        """
        return Asset info
        :return: Asset info with name && balance
        """
        return json.dumps(self.description()._asdict())

    def __repr__(self):
        """
        return Asset info
        :return: Asset info with name && balance
        """
        return f"{self.balance} {self.name}"

    def add(self, amount=Decimal(0)):
        """
        add amount to balance
        :param amount: amount to add
        :type amount: Decimal
        :return: entity itself
        :rtype: BrokerAsset
        """
        self.balance += amount
        return self

    def sub(self, amount=Decimal(0), allow_negative_balance=False):
        """
        subtract amount from balance. if balance is not enough, an error will be raised.

        :param amount: amount to subtract
        :type amount: Decimal
        :param allow_negative_balance: allow balance is negative
        :type allow_negative_balance: bool
        :return:
        :rtype:
        """
        base = self.balance if self.balance != Decimal(0) else Decimal(amount)

        if base == Decimal(0):  # amount and balance is both 0
            return self
        if allow_negative_balance:
            self.balance -= amount
        else:
            # if difference between amount and balance is below 0.01%, will deduct all the balance
            # That's because, the amount calculated by v3_core, has some acceptable error.
            if abs((self.balance - amount) / base) < 0.00001:
                self.balance = Decimal(0)
            elif self.balance - amount < Decimal(0):
                raise AssertionError(
                    f"insufficient balance, balance is {self.balance}{self.name}, "
                    f"but sub amount is {amount}{self.name}"
                )
            else:
                self.balance -= amount

        return self

    def amount_in_wei(self):
        """
        return balance ** decimal

        :return: self.balance * 10 ** self.decimal
        """
        return self.balance * Decimal(10**self.decimal)

    def description(self) -> AssetDescription:
        return AssetDescription(self.name, float(self.balance))


class ActionTypeEnum(Enum):
    """
    Trade types

    * add_liquidity,
    * remove_liquidity,
    * buy,
    * sell,
    * collect_fee
    """

    uni_lp_add_liquidity = "add_liquidity"
    uni_lp_remove_liquidity = "remove_liquidity"
    uni_lp_buy = "buy"
    uni_lp_sell = "sell"
    uni_lp_swap = "swap"
    uni_lp_collect = "collect"
    aave_supply = "supply"
    aave_withdraw = "withdraw"
    aave_borrow = "borrow"
    aave_repay = "repay"
    aave_liquidation = "liquidation"
    option_buy = "option_buy"
    option_sell = "option_sell"
    option_exercise = "option_exercise"
    squeeth_open_vault = "open_vault"
    squeeth_update_collateral = "update_collateral"
    squeeth_update_short = "update_short"
    squeeth_deposit_lp = "deposit_uni_lp"
    squeeth_withdraw_lp = "withdraw_uni_lp"
    squeeth_reduce_debt = "reduce_debt"
    squeeth_liquidation = "liquidation"

    def __str__(self):
        return self.name

    def __repr__(self):
        return self.name


@dataclass
class BaseAction(object):
    """
    Parent class of broker actions,

    :param market: market
    :type market: MarketInfo
    :param action_type: action type
    :type action_type: ActionTypeEnum
    :param timestamp: action time
    :type timestamp: datetime

    """

    market: MarketInfo
    action_type: ActionTypeEnum = field(default=False, init=False)
    timestamp: datetime = field(default=False, init=False)

    def get_output_str(self):
        return str(self)

    def set_type(self):
        pass


@dataclass
class MarketBalance:
    """
    MarketBalance properties

    :param net_value: total net value in this market(in base token)
    :type net_value: Decimal
    """

    net_value: Decimal


@dataclass
class AccountStatusCommon:
    """
    AccountStatusCommon properties

    :type timestamp: datetime
    :type net_value: Decimal, default 0
    """

    timestamp: datetime
    net_value: Decimal = Decimal(0)


@dataclass
class MarketStatus:
    """
    MarketStatus properties

    :type timestamp: datetime
    """

    timestamp: datetime | None
    data: pd.Series | None = None


T = TypeVar("T")


class MarketDict(Generic[T]):
    """
    Market Dict with get/set function
    """

    def __init__(self):
        self.data: Dict[MarketInfo, T] = {}
        self._default: MarketInfo | None = None

    def __getitem__(self, item) -> T:
        return self.data[item]

    def __setitem__(self, key: MarketInfo, value: T):
        if len(self.data) == 0:
            self._default = key
        self.data[key] = value
        setattr(self, key.name, value)

    @property
    def default(self) -> T:
        """
        get default value in MarketDict
        :return:
        """
        return self.data[self._default]

    def get_default_key(self):
        """
        get default key
        :return:
        """
        return self._default

    def set_default_key(self, value: MarketInfo):
        """
        set default key
        :param value:
        :return:
        """
        self._default = value

    def items(self) -> (List[MarketInfo], List[T]):
        """
        get dict items
        :return:
        """
        return self.data.items()

    def keys(self) -> List[MarketInfo]:
        """
        get dict keys
        :return:
        """
        return self.data.keys()

    def values(self) -> List[T]:
        """
        get dict values
        :return:
        """
        return self.data.values()

    def __contains__(self, item):
        """
        check if item in dict
        :param item:
        :return:
        """
        return item in self.data

    def __len__(self):
        """
        len of dict
        :return:
        """
        return len(self.data)


class AssetDict(Generic[T]):
    def __init__(self):
        """
        init AssetDict
        """
        self.data: Dict[TokenInfo, T] = {}

    def __getitem__(self, item) -> T:
        """
        get item magic method
        :param item:
        :return:
        """
        return self.data[item]

    def __setitem__(self, key: TokenInfo, value: T):
        """
        set item magic method
        :param key:
        :param value:
        :return:
        """
        self.data[key] = value
        setattr(self, key.name, value)

    def items(self) -> (List[TokenInfo], List[T]):
        """
        get items from dict
        :return:
        """
        return self.data.items()

    def keys(self) -> List[TokenInfo]:
        """
        get keys from dict
        :return:
        """
        return self.data.keys()

    def values(self) -> List[T]:
        """
        get values from dict
        :return:
        """
        return self.data.values()

    def __contains__(self, item):
        """
        check if item in dict
        :param item:
        :return:
        """
        return item in self.data

    def __len__(self):
        """
        length of dict
        :return:
        """
        return len(self.data)


@dataclass
class AccountStatus(AccountStatusCommon):
    """
    Account Status
    :param asset_balances: balance of asset
    :param market_status:
    """

    asset_balances: AssetDict[Decimal] = field(default_factory=AssetDict)
    market_status: MarketDict[MarketBalance] = field(default_factory=MarketDict)

    def to_array(self) -> List:
        """
        market_status value to list
        :return:
        """
        result = [self.net_value]
        for balance in self.asset_balances.values():
            result.append(balance)
        for market, status in self.market_status.items():
            for k, v in vars(status).items():
                result.append(v)
        return result

    def get_names(self) -> List:
        """
        get market_status market name
        :return:
        """
        result = ["net_value"]
        for asset in self.asset_balances.keys():
            result.append(asset.name)
        for market, status in self.market_status.items():
            base_name = market.name
            for k, v in vars(status).items():
                result.append(f"{base_name}_{k}")
        return result

    @staticmethod
    def to_dataframe(status_list: []) -> pd.DataFrame:
        """
        status list convert to dataframe
        :param status_list:
        :return:
        """
        index = [i.timestamp for i in status_list]
        if len(index) > 0:
            return pd.DataFrame(
                columns=status_list[0].get_names(), index=index, data=map(lambda d: d.to_array(), status_list)
            )
        else:
            return pd.DataFrame()


class PositionManager:
    def __init__(self):
        self.positions: Dict[str, Decimal] = {}
        self.keys: List[T] = []

    def add(self, stock: T, amount: Decimal) -> Decimal:
        key = str(stock)
        if key not in self.positions:
            self.positions[key] = Decimal(0)
            self.keys.append(stock)
        self.positions[key] += amount
        return self.positions[key]

    def substract(self, stock: T, amount: Decimal) -> Decimal:
        key = str(stock)
        if key not in self.positions:
            raise DemeterError(f"{stock} not in get_position")
        if self.positions[key] < amount:
            raise DemeterError(f"insufficient amount for {key}")
        self.positions[key] -= amount
        return self.positions[key]

    def get(self, stock: T) -> Decimal:
        key = str(stock)
        if key not in self.positions:
            raise DemeterError(f"{key} not in get_position")
        return self.positions[key]

    def has(self, stock: T) -> bool:
        key = str(stock)
        return key in self.positions


@dataclass
class RowData:
    timestamp: datetime  # Time of this iteration
    row_id: int  # index of this iteration, start from 0
    prices: pd.Series  # price of tokens at this time
    market_status: MarketDict[Union[pd.Series, pd.DataFrame]] = MarketDict()  # status of markets at this time


BASE_FREQ = "1min"



from dataclasses import dataclass
from datetime import datetime, timedelta
from typing import Callable, Any, List

import pandas as pd

from .. import RowData
from .._typing import DemeterError


def to_minute(time: datetime) -> datetime:
    """
    Convert a datetime instance to minute(Just set its second to 0)

    :param time: time instance to convert
    :type time: datetime
    :return: time whose second is zero
    :rtype: datetime
    """
    return datetime(time.year, time.month, time.day, time.hour, time.minute)


class Trigger:
    """
    Abstract trigger.

    Trigger will do something(decide by do function) when condition is met(decide by when function)

    Extra args can be passed through kwargs. e.g. tg = Trigger(lambda r:r, extra_arg1=0, extra_arg2="1")

    :param do: which action to take.
    :type do: Callable[[RowData], Any]
    """

    def __init__(self, do: Callable[[RowData], Any], **kwargs):
        self._do = do if do is not None else lambda x: x
        self.kwargs = kwargs

    def when(self, row_data: RowData) -> bool:
        """
        If the condition is met or not.

        :param row_data: data of this iteration. used to decide to if the condition is meet.
        :type row_data: RowData
        :return: if condition is met, return true
        :rtype: bool
        """
        return False

    def do(self, row_data: RowData):
        """
        when condition is met, what actions will be taken

        :param row_data: data of this iteration
        :type row_data: condition
        :return: Anything do function returns
        :rtype: Any
        """
        return self._do(row_data, **self.kwargs)


class AtTimeTrigger(Trigger):
    """
    Trigger action at a specific time

    Extra args can be passed through kwargs.

    :param time: time to trigger action
    :type time: datetime
    :param do: which action to take.
    :type do: Callable[[RowData], Any]
    """

    def __init__(self, time: datetime, do, **kwargs):
        self._time = to_minute(time)
        super().__init__(do, **kwargs)

    def when(self, row_data: RowData) -> bool:
        return row_data.timestamp == self._time


class AtTimesTrigger(Trigger):
    """
    trigger action at some specific times

    Extra args can be passed through kwargs.

    :param time: when current timestamp is in List[datetime], will trigger action
    :type time: List[datetime]
    :param do: which action to take.
    :type do: Callable[[RowData], Any]

    """

    def __init__(self, time: List[datetime], do, **kwargs):
        self._time = [to_minute(t) for t in time]
        super().__init__(do, **kwargs)

    def when(self, row_data: RowData) -> bool:
        return self._time in row_data.timestamp


@dataclass
class TimeRange:
    """
    Time range
    """

    start: datetime
    end: datetime


class TimeRangeTrigger(Trigger):
    """
    trigger action at a time range

    Extra args can be passed through kwargs. e.g. tg = Trigger(lambda r:r, extra_arg1=0, extra_arg2="1")

    :param time_range: when current timestamp is between time range, will trigger action, end time will not be included
    :type time_range: TimeRange
    :param do: which action to take.
    :type do: Callable[[RowData], Any]
    """

    def __init__(self, time_range: TimeRange, do, **kwargs):
        self._time_range = TimeRange(to_minute(time_range.start), to_minute(time_range.end))
        super().__init__(do, **kwargs)

    def when(self, row_data: RowData) -> bool:
        return self._time_range.start <= row_data.timestamp < self._time_range.end


class TimeRangesTrigger(Trigger):
    """
    trigger action at some time range

    Extra args can be passed through kwargs. e.g. tg = Trigger(lambda r:r, extra_arg1=0, extra_arg2="1")

    :param time_range: when current timestamp is between any time range, will trigger action, end time will not be included
    :type time_range: List[TimeRange]
    :param do: which action to take.
    :type do: Callable[[RowData], Any]
    """

    def __init__(self, time_range: List[TimeRange], do, **kwargs):
        self._time_range: [TimeRange] = [TimeRange(to_minute(t.start), to_minute(t.end)) for t in time_range]
        super().__init__(do, **kwargs)

    def when(self, row_data: RowData) -> bool:
        for r in self._time_range:
            if r.start <= row_data.timestamp < r.end:
                return True
        return False


def _check_time_delta(delta: timedelta):
    if delta.total_seconds() % 60 != 0:
        raise DemeterError("min time span is 1 minute")


class PeriodTrigger(Trigger):
    """
    Trigger action periodically

    Extra args can be passed through kwargs. e.g. tg = Trigger(lambda r:r, extra_arg1=0, extra_arg2="1")

    :param time_delta: Period
    :type time_delta: timedelta
    :param do: which action to take.
    :type do: Callable[[RowData], Any]
    :param trigger_immediately: whither to trigger action when back test just started
    :type trigger_immediately: bool
    """

    def __init__(self, time_delta: timedelta, do, trigger_immediately=False, **kwargs):
        self._next_match = None
        self._delta = time_delta
        self._trigger_immediately = trigger_immediately
        _check_time_delta(time_delta)
        super().__init__(do, **kwargs)

    def reset(self):
        self._next_match = None

    def when(self, row_data: RowData) -> bool:
        if self._next_match is None:
            self._next_match = row_data.timestamp + self._delta
            return self._trigger_immediately

        if self._next_match == row_data.timestamp:
            self._next_match = self._next_match + self._delta
            return True

        return False


class PeriodsTrigger(Trigger):
    """
    trigger action periodically, but you can set multiple period.

    Extra args can be passed through kwargs. e.g. tg = Trigger(lambda r:r, extra_arg1=0, extra_arg2="1")

    :param time_delta: Periods,
    :type time_delta: List[timedelta]
    :param do: which action to take.
    :type do: Callable[[RowData], Any]
    :param trigger_immediately: whither to trigger action when back test just started
    :type trigger_immediately: bool
    """

    def __init__(self, time_delta: List[timedelta], do, trigger_immediately=False, **kwargs):
        self._next_matches = [None for _ in time_delta]
        self._deltas = time_delta
        self._trigger_immediately = trigger_immediately

        for td in time_delta:
            _check_time_delta(td)
        super().__init__(do, **kwargs)

    def reset(self):
        self._next_matches = [None for _ in self._deltas]

    def when(self, row_data: RowData) -> bool:
        if self._next_matches[0] is None:
            self._next_matches = [row_data.timestamp + d for d in self._deltas]
            return self._trigger_immediately

        for i in range(len(self._deltas)):
            if self._next_matches[i] == row_data.timestamp:
                self._next_matches[i] = self._next_matches[i] + self._deltas[i]
                return True

        return False


class PriceTrigger(Trigger):
    """
    Trigger when price meet a customized condition

    Extra args can be passed through kwargs. e.g. tg = Trigger(lambda r:r, extra_arg1=0, extra_arg2="1")

    :param condition: customized condition, arg is price of tokens
    :type condition: Callable[[pd.Series], bool]
    :param do: which action to take.
    :type do: Callable[[RowData], Any]
    """

    def __init__(self, condition: Callable[[pd.Series], bool], do, **kwargs):
        self._condition = condition
        super().__init__(do, **kwargs)

    def when(self, row_data: RowData) -> bool:
        return self._condition(row_data.prices)


class CustomizedTrigger(Trigger):
    """
    Trigger on customized condition

    Extra args can be passed through kwargs. e.g. tg = Trigger(lambda r:r, extra_arg1=0, extra_arg2="1")

    :param condition: customized condition
    :type condition: Callable[[RowData], bool]
    :param do: which action to take.
    :type do: Callable[[RowData], Any]
    """

    def __init__(self, condition: Callable[[RowData], bool], do, **kwargs):
        self._condition = condition
        super().__init__(do, **kwargs)

    def when(self, row_data: RowData) -> bool:
        return self._condition(row_data)



from typing import Dict, List

import pandas as pd

from .._typing import UnitDecimal, DemeterError, EvaluatorEnum
from ..broker import AccountStatus, AccountStatusCommon
from ..broker._typing import MarketDict
from .math_helper import max_draw_down_fast, annualized_returns, get_benchmark_returns


class Evaluator(object):
    """
    Calculate evaluator indicator for strategy.

    :param init_status: Status before back test start
    :type init_status: AccountStatus
    :param data: account status in dataframe format
    :type data: DataFrame
    :param prices: dataframe of AccountStatusCommon
    :type prices: DataFrame
    :param actions: all actions during backtest
    :type actions: List
    """

    def __init__(self, init_status: AccountStatus, data: pd.DataFrame, prices: pd.DataFrame, markets: MarketDict):
        """
        init Evaluator

        """
        self.init_status: AccountStatus = init_status
        self.end_status: AccountStatusCommon = data.iloc[-1]
        self.prices: pd.DataFrame = prices
        self.data: pd.DataFrame = data
        self.markets = markets
        if len(data) < 2:
            raise DemeterError("not enough data")
        self.time_span_in_day = len(data.index) * (data.index[1] - data.index[0]).seconds / (60 * 60 * 24)
        self._result = None

    def run(self, enables: List[EvaluatorEnum]) -> Dict[EvaluatorEnum, UnitDecimal]:
        """
        run evaluator

        :param enables: which evaluator to enable
        :type enables: List[EvaluatorEnum]
        :return: dict of results, key is evaluator, value is result of evaluator
        :rtype: Dict[EvaluatorEnum, UnitDecimal]
        """
        if EvaluatorEnum.all in enables:
            enables = [x for x in EvaluatorEnum]
            enables = filter(lambda x: x.value > 0, enables)
        result_dict: Dict[EvaluatorEnum, UnitDecimal] = {}
        for request in enables:
            match request:
                case EvaluatorEnum.annualized_returns:
                    result = UnitDecimal(
                        annualized_returns(
                            self.init_status.net_value,
                            self.end_status.net_value,
                            self.time_span_in_day,
                        ),
                        "",
                    )
                case EvaluatorEnum.benchmark_returns:
                    result = UnitDecimal(
                        get_benchmark_returns(
                            self.init_status.net_value,
                            self.prices.iloc[0],
                            self.prices.iloc[-1],
                            self.time_span_in_day,
                        ),
                        "",
                    )
                case EvaluatorEnum.max_draw_down:
                    result = UnitDecimal(max_draw_down_fast(self.data.net_value), "")
                case EvaluatorEnum.net_value:
                    result = UnitDecimal(self.end_status.net_value / self.init_status.net_value)
                case EvaluatorEnum.profit:
                    result = UnitDecimal(self.end_status.net_value - self.init_status.net_value)
                case EvaluatorEnum.net_value_up_down_rate:
                    result = UnitDecimal((self.end_status.net_value - self.init_status.net_value) /
                                         self.init_status.net_value)
                case EvaluatorEnum.eth_up_down_rate:
                    result = UnitDecimal((self.prices.iloc[-1]["ETH"] - self.prices.iloc[0]["ETH"]) /
                                         self.prices.iloc[0]["ETH"])
                case EvaluatorEnum.position_fee_profit:
                    fee_value = UnitDecimal(0)
                    for _, market in self.markets.items():
                        fee_df = self.data[[f"{market.market_info.name}_base_uncollected",
                                            f"{market.market_info.name}_quote_uncollected"]]
                        fee_df = fee_df.sort_values(by=[f"{market.market_info.name}_base_uncollected",
                                                    f"{market.market_info.name}_quote_uncollected"],
                                                    ascending=[False, False])
                        fee_price_df = pd.merge(fee_df, self.prices, how="left", left_index=True, right_index=True)
                        latest_fee = fee_price_df.iloc[0]
                        fee_value += (latest_fee[f"{market.market_info.name}_base_uncollected"] +
                                      latest_fee[f"{market.market_info.name}_quote_uncollected"] * latest_fee["ETH"])
                    result = UnitDecimal(fee_value)
                case EvaluatorEnum.position_fee_annualized_returns:
                    fee_value = UnitDecimal(0)
                    for _, market in self.markets.items():
                        fee_df = self.data[(self.data[f"{market.market_info.name}_base_uncollected"] > 0) |
                                           (self.data[f"{market.market_info.name}_quote_uncollected"] > 0)]
                        fee_df = fee_df[[f"{market.market_info.name}_base_uncollected",
                                         f"{market.market_info.name}_quote_uncollected"]]
                        if not fee_df.empty:
                            fee_df = fee_df.sort_values(by=[f"{market.market_info.name}_base_uncollected",
                                                            f"{market.market_info.name}_quote_uncollected"],
                                                        ascending=[False, False])
                            fee_price_df = pd.merge(fee_df, self.prices, how="left", left_index=True, right_index=True)
                            latest_fee = fee_price_df.iloc[0]
                            fee_value = (latest_fee[f"{market.market_info.name}_base_uncollected"] +
                                         latest_fee[f"{market.market_info.name}_quote_uncollected"] * latest_fee["ETH"])
                            fee_annualized_returns = ((fee_value / self.init_status.net_value) *
                                                      UnitDecimal(len(fee_df) / len(self.data) * 365))
                            fee_value += UnitDecimal(fee_annualized_returns)
                        else:
                            fee_value += UnitDecimal(0)
                    result = fee_value
                case EvaluatorEnum.position_market_time_rate:
                    market_time_rate = UnitDecimal(0)
                    for _, market in self.markets.items():
                        fee_df = self.data[(self.data[f"{market.market_info.name}_base_uncollected"] > 0) |
                                           (self.data[f"{market.market_info.name}_quote_uncollected"] > 0)]
                        market_time_rate += UnitDecimal(len(fee_df) / len(self.data))
                    result = market_time_rate
                case _:
                    raise DemeterError(f"{request} has not implied")
            result_dict[request] = result
        self._result = result_dict
        return result_dict

    @property
    def result(self) -> Dict[EvaluatorEnum, UnitDecimal]:
        """
        Return evaluate result after run()

        :return: evaluate result
        :rtype: Dict[EvaluatorEnum, UnitDecimal]
        """
        return self._result

    def __str__(self):
        """
        Evaluator print function
        """
        str_array = []
        for k, v in self._result.items():
            str_array.append(f"{k.name:<35}:{v:.15}")
        return ";\n".join(str_array)



from decimal import Decimal

# -*- coding: utf-8 -*-
"""
!!! IMPORTANT 

this module is developed and enhanced from active-strategy-framework of GammaStrategies
source code: https://github.com/GammaStrategies/active-strategy-framework/blob/main/UNI_v3_funcs.py

Original author information: 
=============================================

Created on Mon Jun 14 18:53:09 2021

@author: JNP
"""

"""liquitidymath"""
"""Python library to emulate the calculations done in liquiditymath.sol of UNI_V3 peryphery contract"""

# sqrtP: format X96 = int(1.0001**(tick/2)*(2**96))
# liquidity: int
# sqrtA = price for lower tick
# sqrtB = price for upper tick
"""get_amounts function"""


# Use 'get_amounts' function to calculate amounts as a function of liquitidy and price range
def get_amount0(sqrtA: int, sqrtB: int, liquidity: int, decimals: int) -> Decimal:
    if sqrtA > sqrtB:
        (sqrtA, sqrtB) = (sqrtB, sqrtA)
    amount0 = (Decimal(liquidity * 2**96 * (sqrtB - sqrtA)) / sqrtB / sqrtA) / 10**decimals
    return amount0


def get_amount1(sqrtA: int, sqrtB: int, liquidity: int, decimals: int) -> Decimal:
    if sqrtA > sqrtB:
        (sqrtA, sqrtB) = (sqrtB, sqrtA)
    amount1 = Decimal(liquidity * (sqrtB - sqrtA)) / 2**96 / 10**decimals
    return amount1


def get_sqrt(tick: int):
    return Decimal(1.0001 ** (tick / 2) * (2**96))


def get_amounts(
    sqrt_price_x96: int,
    tickA: int,
    tickB: int,
    liquidity: int,
    decimal0: int,
    decimal1: int,
) -> (Decimal, Decimal):
    sqrt = sqrt_price_x96
    sqrtA = get_sqrt_ratio_at_tick(tickA)
    sqrtB = get_sqrt_ratio_at_tick(tickB)

    if sqrtA > sqrtB:
        (sqrtA, sqrtB) = (sqrtB, sqrtA)

    if sqrt <= sqrtA:
        amount0 = get_amount0(sqrtA, sqrtB, liquidity, decimal0)
        return amount0, Decimal(0)
    elif sqrtB > sqrt > sqrtA:
        amount0 = get_amount0(sqrt, sqrtB, liquidity, decimal0)
        amount1 = get_amount1(sqrtA, sqrt, liquidity, decimal1)
        return amount0, amount1
    else:
        amount1 = get_amount1(sqrtA, sqrtB, liquidity, decimal1)
        return Decimal(0), amount1


"""get token amounts relation"""


# Use this formula to calculate amount of t0 based on amount of t1 (required before calculate liquidity)
# relation = t1/t0
def amounts_relation(tick: int, tickA: int, tickB: int, decimals0: int, decimals1: int) -> Decimal:
    sqrt = (1.0001**tick / 10 ** (decimals1 - decimals0)) ** (1 / 2)
    sqrtA = (1.0001**tickA / 10 ** (decimals1 - decimals0)) ** (1 / 2)
    sqrtB = (1.0001**tickB / 10 ** (decimals1 - decimals0)) ** (1 / 2)

    if sqrt == sqrtA or sqrt == sqrtB:
        relation = 0

    relation = (sqrt - sqrtA) / ((1 / sqrt) - (1 / sqrtB))
    return relation


"""get_liquidity function"""


def mul_div(a: int, b: int, denominator: int) -> int:
    """
    this function is very long in contract. but It's because max length in solidity is limited.

    But python has unlimit integer.

    ensure all the parameter is int !
    """
    return a * b // denominator


def get_liquidity_for_amount0(sqrtA: int, sqrtB: int, amount: int) -> int:
    if sqrtA > sqrtB:
        (sqrtA, sqrtB) = (sqrtB, sqrtA)

    intermediate = mul_div(sqrtA, sqrtB, 2**96)
    return mul_div(amount, intermediate, sqrtB - sqrtA)


def get_liquidity_for_amount1(sqrtA: int, sqrtB: int, amount: int) -> int:
    if sqrtA > sqrtB:
        (sqrtA, sqrtB) = (sqrtB, sqrtA)

    return mul_div(amount, 2**96, sqrtB - sqrtA)


def to_wei(amount, decimals) -> int:
    return int(amount * 10**decimals)


def get_liquidity(
    sqrt_price_x96: int,
    tickA: int,
    tickB: int,
    amount0: Decimal,
    amount1: Decimal,
    decimal0: int,
    decimal1: int,
) -> int:
    sqrt = sqrt_price_x96
    sqrtA = get_sqrt_ratio_at_tick(tickA)
    sqrtB = get_sqrt_ratio_at_tick(tickB)

    if sqrtA > sqrtB:
        (sqrtA, sqrtB) = (sqrtB, sqrtA)
    amount0wei: int = to_wei(amount0, decimal0)
    amount1wei: int = to_wei(amount1, decimal1)
    if sqrt <= sqrtA:
        liquidity0 = get_liquidity_for_amount0(sqrtA, sqrtB, amount0wei)
        return liquidity0
    elif sqrtB > sqrt > sqrtA:
        liquidity0 = get_liquidity_for_amount0(sqrt, sqrtB, amount0wei)
        liquidity1 = get_liquidity_for_amount1(sqrtA, sqrt, amount1wei)
        liquidity = liquidity0 if liquidity0 < liquidity1 else liquidity1
        return liquidity
    else:
        liquidity1 = get_liquidity_for_amount1(sqrtA, sqrtB, amount1wei)
        return liquidity1


def get_sqrt_ratio_at_tick(tick: int) -> int:
    """
    (sqrt(1.0001) ** tick) * (2**96)
    """
    tick = int(tick)
    abs_tick = tick if tick >= 0 else -tick
    assert abs_tick <= 887272

    # Those magic number stands for 1/sqrt(1.0001)^1, 1/sqrt(1.0001)^2, 1/sqrt(1.0001)^4....
    ratio: int = 0xFFFCB933BD6FAD37AA2D162D1A594001 if abs_tick & 0x1 != 0 else 0x100000000000000000000000000000000
    if abs_tick & 0x2 != 0:
        ratio = (ratio * 0xFFF97272373D413259A46990580E213A) >> 128
    if abs_tick & 0x4 != 0:
        ratio = (ratio * 0xFFF2E50F5F656932EF12357CF3C7FDCC) >> 128
    if abs_tick & 0x8 != 0:
        ratio = (ratio * 0xFFE5CACA7E10E4E61C3624EAA0941CD0) >> 128
    if abs_tick & 0x10 != 0:
        ratio = (ratio * 0xFFCB9843D60F6159C9DB58835C926644) >> 128
    if abs_tick & 0x20 != 0:
        ratio = (ratio * 0xFF973B41FA98C081472E6896DFB254C0) >> 128
    if abs_tick & 0x40 != 0:
        ratio = (ratio * 0xFF2EA16466C96A3843EC78B326B52861) >> 128
    if abs_tick & 0x80 != 0:
        ratio = (ratio * 0xFE5DEE046A99A2A811C461F1969C3053) >> 128
    if abs_tick & 0x100 != 0:
        ratio = (ratio * 0xFCBE86C7900A88AEDCFFC83B479AA3A4) >> 128
    if abs_tick & 0x200 != 0:
        ratio = (ratio * 0xF987A7253AC413176F2B074CF7815E54) >> 128
    if abs_tick & 0x400 != 0:
        ratio = (ratio * 0xF3392B0822B70005940C7A398E4B70F3) >> 128
    if abs_tick & 0x800 != 0:
        ratio = (ratio * 0xE7159475A2C29B7443B29C7FA6E889D9) >> 128
    if abs_tick & 0x1000 != 0:
        ratio = (ratio * 0xD097F3BDFD2022B8845AD8F792AA5825) >> 128
    if abs_tick & 0x2000 != 0:
        ratio = (ratio * 0xA9F746462D870FDF8A65DC1F90E061E5) >> 128
    if abs_tick & 0x4000 != 0:
        ratio = (ratio * 0x70D869A156D2A1B890BB3DF62BAF32F7) >> 128
    if abs_tick & 0x8000 != 0:
        ratio = (ratio * 0x31BE135F97D08FD981231505542FCFA6) >> 128
    if abs_tick & 0x10000 != 0:
        ratio = (ratio * 0x9AA508B5B7A84E1C677DE54F3E99BC9) >> 128
    if abs_tick & 0x20000 != 0:
        ratio = (ratio * 0x5D6AF8DEDB81196699C329225EE604) >> 128
    if abs_tick & 0x40000 != 0:
        ratio = (ratio * 0x2216E584F5FA1EA926041BEDFE98) >> 128
    if abs_tick & 0x80000 != 0:
        ratio = (ratio * 0x48A170391F7DC42444E8FA2) >> 128

    if tick > 0:
        # type(uint256).max
        ratio = int(115792089237316195423570985008687907853269984665640564039457584007913129639935 // ratio)

    # this divides by 1<<32 rounding up to go from a Q128.128 to a Q128.96.
    # we then downcast because we know the result always fits within 160 bits due to our tick input constraint
    # we round up in the division so getTickAtSqrtRatio of the output price is always consistent
    sqrt_price_x96 = (ratio >> 32) + (0 if ratio % (1 << 32) == 0 else 1)
    return sqrt_price_x96



from datetime import timedelta

import pandas as pd
from pandas import Timedelta

from .._typing import DemeterError


def get_real_n(data: pd.Series, window: timedelta):
    """
    Get how many rows of data should be included in time window

    :param data: pd.Series whose timestamp interval are same.
    :type data: Series
    :param window: time window
    :param window: timedelta
    :return: row count of corresponding time window
    :rtype: int
    """
    if data.size < 2:
        raise DemeterError("not enough data for simple_moving_average")
    timespan: Timedelta = data.index[1] - data.index[0]
    if timespan.seconds % 60 != 0:
        return DemeterError("no seconds is allowed")
    data_span_in_minute = timespan.total_seconds() / 60
    window_minutes = int(window.total_seconds()) // 60
    if window_minutes % data_span_in_minute != 0:
        raise DemeterError(f"window span is {window_minutes}minutes, " f"but data span is {data_span_in_minute}minute, cannot divide exactly")
    real_n = window_minutes // int(data_span_in_minute)
    if data.size < real_n:
        raise DemeterError("not enough data for simple_moving_average")
    return real_n



from decimal import Decimal

from ._typing import UniV3Pool, Position, UniV3PoolStatus, PositionInfo
from .helper import base_unit_price_to_tick, from_atomic_unit
from .liquitidy_math import get_amounts, get_liquidity


class V3CoreLib(object):
    @staticmethod
    def new_position(
        pool: UniV3Pool,
        token0_amount: Decimal,
        token1_amount: Decimal,
        lower_tick: int,
        upper_tick: int,
        sqrt_price_x96: int,
    ):
        """
        create new get_position

        :param pool: operation on which pool
        :param token0_amount: token0 amount
        :param token1_amount: token1 amount
        :param lower_tick: lower tick
        :param upper_tick: upper tick
        :param sqrt_price_x96: sqrt(price) * 2^96
        :return: token0 get_position, token1 get_position, liquid, get_position instance
        """
        position_liq = get_liquidity(
            sqrt_price_x96,
            lower_tick,
            upper_tick,
            token0_amount,
            token1_amount,
            pool.token0.decimal,
            pool.token1.decimal,
        )
        token0_in_position, token1_in_position = get_amounts(
            sqrt_price_x96,
            lower_tick,
            upper_tick,
            position_liq,
            pool.token0.decimal,
            pool.token1.decimal,
        )
        new_position_entity = PositionInfo(lower_tick=lower_tick, upper_tick=upper_tick)
        return (
            token0_in_position,
            token1_in_position,
            int(position_liq),
            new_position_entity,
        )

    @staticmethod
    def close_position(pool: UniV3Pool, position_info: PositionInfo, liquidity, sqrt_price_x96):
        """
        close get_position

        :param pool: operation on which pool
        :param position_info: get_position info
        :param liquidity: liquidity
        :param sqrt_price_x96: sqrt(price) * 2^96
        :return: token amount
        """
        return V3CoreLib.get_token_amounts(pool, position_info, sqrt_price_x96, liquidity)

    @staticmethod
    def get_token_amounts(pool: UniV3Pool, pos: PositionInfo, sqrt_price_x96, liquidity) -> (Decimal, Decimal):
        """
        get token amount in get_position

        :param pool: operation on which pool
        :param pos: get_position info
        :param sqrt_price_x96: sqrt(price) * 2^96
        :param liquidity: liquidity
        :return: token0 amount, token1 amount
        """
        if liquidity == 0:  # performance improve
            return 0, 0
        amount0, amount1 = get_amounts(
            sqrt_price_x96,
            pos.lower_tick,
            pos.upper_tick,
            liquidity,
            pool.token0.decimal,
            pool.token1.decimal,
        )
        return amount0, amount1

    @staticmethod
    def quote_price_pair_to_tick(pool: UniV3Pool, lower_quote_price: Decimal, upper_quote_price: Decimal):
        """
        quote price pair to tick

        :param pool: operation on which pool
        :param lower_quote_price: lower quote price
        :param upper_quote_price: upper quote price
        :return: lower_tick, upper_tick
        """
        lower_tick = base_unit_price_to_tick(
            lower_quote_price,
            pool.token0.decimal,
            pool.token1.decimal,
            pool.is_token0_quote,
        )
        upper_tick = base_unit_price_to_tick(
            upper_quote_price,
            pool.token0.decimal,
            pool.token1.decimal,
            pool.is_token0_quote,
        )
        lower_tick, upper_tick = (lower_tick, upper_tick) if not pool.is_token0_quote else (upper_tick,lower_tick)
        return lower_tick, upper_tick

    @staticmethod
    def update_fee(last_tick:int,pool: UniV3Pool, pos: PositionInfo, position: Position, state: UniV3PoolStatus):
        """
        update fee

        :param pool: operation on which pool
        :param pos: get_position info
        :param position: get_position
        :param state: UniV3PoolStatus
        :param last_tick:
        :return: None
        """

        # in most cases, tick will not cross to on_bar one, which means L will not change.
        def calc_amounts():
            share = Decimal(position.liquidity) / Decimal(state.currentLiquidity)
            position.pending_amount0 += from_atomic_unit(state.inAmount0, pool.token0.decimal) * share * pool.fee_rate
            position.pending_amount1 += from_atomic_unit(state.inAmount1, pool.token1.decimal) * share * pool.fee_rate

        condition_in_position = pos.upper_tick >= state.closeTick >= pos.lower_tick
        if last_tick:
            condition_over_position = (last_tick > pos.upper_tick and state.closeTick < pos.lower_tick) or (
                    state.closeTick > pos.upper_tick and last_tick < pos.lower_tick
            )
            condition_in_to_out_position = pos.upper_tick >= last_tick >= pos.lower_tick and (
                    state.closeTick > pos.upper_tick or state.closeTick < pos.lower_tick
            )
        else:
            condition_in_to_out_position = condition_over_position = False

        if condition_in_position or condition_over_position or condition_in_to_out_position:
            calc_amounts()



import json
from decimal import Decimal
from enum import Enum
from functools import wraps
from types import SimpleNamespace
from typing import Any, Dict

OUTPUT_WIDTH = 30


def to_decimal(value: Any) -> Decimal:
    """
    convert value to decimal

    :param value: any value
    :type value: Any
    :return: Decimal value
    :rtype: Decimal

    """
    return Decimal(str(value))


def object_to_decimal(num: Any) -> Any:
    """
    If number is float or int, return Decimal, else return original value

    :param value: any value
    :type value: Any
    :return: Decimal value
    :rtype: Any
    """
    return Decimal(str(num)) if (isinstance(num, float) or type(num) == int) else num


def dict_to_object(dict_entity: Dict) -> Any:
    """
    convert dict to object via json

    :param dict_entity: dict instance
    :type dict_entity: Dict
    :return: object
    :rtype: Any
    """
    return json.loads(json.dumps(dict_entity), object_hook=lambda d: SimpleNamespace(**d))


def float_param_formatter(func):
    """
    decorator to convert param to float

    :param func: any function
    :return: function execute result
    """

    @wraps(func)
    def wrapper_func(*args, **kwargs):
        new_args = ()
        for arg in args:
            new_args += (object_to_decimal(arg),)
        for k, v in kwargs.items():
            kwargs[k] = object_to_decimal(v)
        return func(*new_args, **kwargs)

    return wrapper_func


def get_enum_by_name(me: Enum, name: str):
    """
    get enum by name

    :param me: enum class
    :param name: enum item name
    :return: get value in enum
    """
    for e in me:
        if e.name.lower() == name.lower():
            return e
    raise RuntimeError(f"cannot found {name} in {me}, allow value is " + str([x.name for x in me]))


def require(condition: bool, error_msg: str):
    """
    Checking whether the condition is True, if not, will raise a AssertionError

    :param condition: condition
    :type condition: bool
    :param error_msg: error message contains in AssertionError
    :param error_msg: str
    """
    if not condition:
        raise AssertionError(error_msg)



from datetime import timedelta

import numpy as np
import pandas as pd
from pandas._typing import TimedeltaConvertibleTypes, Axis

from .common import get_real_n


def simple_moving_average(
    data: pd.Series | pd.DataFrame,
    window: timedelta = timedelta(hours=5),
    min_periods: int | None = None,
    center: bool = False,
    win_type: str | None = None,
    on: str | None = None,
    closed: str | None = None,
    method: str = "single",
) -> pd.Series:
    """
    calculate simple moving average, Note: window is based on time span

    docs for other params, see https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.Series.rolling.html

    :param data: data
    :type data: Series
    :param window: window width
    :type window: timedelta
    :return: simple moving average data
    :rtype: Series

    """

    return data.rolling(
        window=get_real_n(data, window),
        min_periods=min_periods,
        center=center,
        win_type=win_type,
        on=on,
        closed=closed,
        method=method,
    ).mean()


def exponential_moving_average(
    data: pd.Series | pd.DataFrame,
    com: float | None = None,
    span: float | None = None,
    halflife: float | TimedeltaConvertibleTypes | None = None,
    alpha: float | None = None,
    min_periods: int | None = 0,
    adjust: bool = True,
    ignore_na: bool = False,
    times: str | np.ndarray | pd.DataFrame | pd.Series | None = None,
    method: str = "single",
):
    """
    calculate exponential moving average, just a shortcut for pandas.evm().mean()

    docs for params, see: https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.Series.ewm.html
    """
    return data.ewm(
        com=com,
        span=span,
        halflife=halflife,
        alpha=alpha,
        min_periods=min_periods,
        adjust=adjust,
        ignore_na=ignore_na,
        times=times,
        method=method,
    ).mean()



from _decimal import Decimal
from typing import Dict

from demeter import DECIMAL_0, TokenInfo
from demeter.aave._typing import SupplyKey, ActionKey, BorrowKey
import pandas as pd


class AaveV3CoreLib(object):
    """
    | The core calculation library for aave v3 market
    | Note: All functions are static
    """

    SECONDS_IN_A_YEAR = 31536000
    """how many seconds in a year"""

    HEALTH_FACTOR_LIQUIDATION_THRESHOLD = Decimal("1")
    """
    if CLOSE_FACTOR_HF_THRESHOLD < health factor <  DEFAULT_LIQUIDATION_CLOSE_FACTOR
    only DEFAULT_LIQUIDATION_CLOSE_FACTOR(50%) will be liquidated,
    otherwise HEALTH_FACTOR_LIQUIDATION_THRESHOLD(100%) will be liquidated
    """

    DEFAULT_LIQUIDATION_CLOSE_FACTOR = Decimal("0.5")

    MAX_LIQUIDATION_CLOSE_FACTOR = Decimal("1")

    CLOSE_FACTOR_HF_THRESHOLD = Decimal("0.95")


    @staticmethod
    def rate_to_apy(rate: Decimal) -> Decimal:
        """
        Convert supply/borrow interest rate(get from ReserveDataUpdated event) to apy(annual interest rate)

        :param rate: interest rate(time unit is second)
        :type rate: Decimal
        :return: apy (time unit is year)
        :rtype: Decimal
        """
        return (1 + rate / AaveV3CoreLib.SECONDS_IN_A_YEAR) ** AaveV3CoreLib.SECONDS_IN_A_YEAR - 1

    @staticmethod
    def get_amount(base_amount: Decimal, liquidity_index: Decimal) -> Decimal:
        """
        | Get amount from base amount.
        | Note: base amount is used in contract. Actual user token balance is base_amount * liquidity_index

        :param base_amount: base token amount
        :type base_amount: Decimal
        :param liquidity_index: token liquidity_index at this moment.
        :type liquidity_index: Decimal
        :return: Actually token amount.
        :rtype: Decimal
        """
        return Decimal(base_amount) * Decimal(liquidity_index)

    @staticmethod
    def get_base_amount(amount: Decimal, liquidity_index: Decimal) -> Decimal:
        """
        | Get base amount from amount
        | Note: base amount is used in contract. Actual user token balance is base_amount * liquidity_index

        :param amount: token amount
        :type amount: Decimal
        :param liquidity_index: token liquidity_index at this moment.
        :type liquidity_index: Decimal
        :return: token base amount.
        :rtype: Decimal
        """
        return amount / liquidity_index

    @staticmethod
    def get_max_borrow_value(
        collaterals: Dict[SupplyKey, Decimal],
        borrows: Dict[BorrowKey, Decimal],
        risk_parameters: pd.DataFrame,
    ) -> Decimal:
        """
        | Calculate the max amount to borrow.
        | Note: As aave web app will multiply 0.99 to result, we followed this practice

        :param collaterals: All collaterals. note: unit of dict value(Decimal) is usd
        :type collaterals: Dict[SupplyKey, Decimal]
        :param borrows: All borrows. note: unit of dict value(Decimal) is usd
        :type borrows: Dict[BorrowKey, Decimal]
        :param risk_parameters: risk_parameters of a chain.
        :type risk_parameters: pd.DataFrame
        :return: borrow amount (in usd)
        :rtype: Decimal
        """
        ltv = AaveV3CoreLib.current_ltv(collaterals, risk_parameters)
        total_borrow = Decimal(sum(borrows.values()))
        total_collateral = Decimal(sum(collaterals.values()))

        return (total_collateral * ltv - total_borrow) * Decimal("0.99")  # because dapp web ui will multiply 0.99

    @staticmethod
    def get_min_withdraw_kept_amount(
        token: TokenInfo,
        collaterals: Dict[SupplyKey, Decimal],
        borrows: Dict[BorrowKey, Decimal],
        risk_parameters: pd.DataFrame,
        price: Decimal,
    ) -> Decimal:
        """
        | Get min collateral token amount to keep health factor above 1

        :param token: which token to calculate
        :type token: TokenInfo
        :param collaterals: collateral values of each token, note: dict value(Decimal) is token value in usd
        :type collaterals: Dict[SupplyKey, Decimal]
        :param borrows: borrow values of each token, note: dict value(Decimal) is token value in usd
        :type borrows: Dict[BorrowKey, Decimal]
        :param risk_parameters: token risk parameters of this chain
        :type risk_parameters: DataFrame
        :param price: current token price
        :type price: Decimal
        :return: min token amount should kept to prevent liquidation
        :rtype: Decimal
        """
        # if token is not collateral token, doesn't need to keep any
        if token not in [k.token for k in collaterals.keys()]:
            return DECIMAL_0
        supplies_liq_threshold = Decimal(0)
        for s, v in collaterals.items():
            if s.token != token:
                supplies_liq_threshold += risk_parameters.loc[s.token.name].liqThereshold * v

        amount = (AaveV3CoreLib.HEALTH_FACTOR_LIQUIDATION_THRESHOLD * Decimal(sum(borrows.values())) - supplies_liq_threshold) / risk_parameters.loc[
            token.name
        ].liqThereshold

        return amount / price

    @staticmethod
    def health_factor(collaterals: Dict[SupplyKey, Decimal], borrows: Dict[BorrowKey, Decimal], risk_parameters) -> Decimal:
        """
        Get current health factor in decimal, e.g.0.8

        :param collaterals: collateral values of each token, note: dict value(Decimal) is token value in usd
        :type collaterals: Dict[SupplyKey, Decimal]
        :param borrows: borrow values of each token, note: dict value(Decimal) is token value in usd
        :type borrows: Dict[BorrowKey, Decimal]
        :param risk_parameters: token risk parameters of this chain
        :type risk_parameters: DataFrame
        :return: health factor
        :rtype: Decimal
        """
        # (all supplies * liqThereshold) / all borrows
        a = Decimal(sum([s * risk_parameters.loc[key.token.name].liqThereshold for key, s in collaterals.items()]))
        b = Decimal(sum(borrows.values()))
        return AaveV3CoreLib.safe_div(a, b)

    @staticmethod
    def current_ltv(collaterals: Dict[SupplyKey, Decimal], risk_parameters) -> Decimal:
        """
        Get max ltv of this user, calculated by token ltv and positions

        :param collaterals: collateral values, note: dict value(Decimal) is token value in usd
        :type collaterals: Dict[SupplyKey, Decimal]
        :param risk_parameters: token risk parameters of this chain
        :type risk_parameters: DataFrame
        :return: max ltv
        :rtype: Decimal
        """
        all_supplies = DECIMAL_0
        for t, s in collaterals.items():
            all_supplies += s * risk_parameters.loc[t.token.name].LTV

        amount = sum(collaterals.values())
        return AaveV3CoreLib.safe_div(all_supplies, Decimal(amount))

    @staticmethod
    def total_liquidation_threshold(collaterals: Dict[SupplyKey, Decimal], risk_parameters):
        """
        | Get total liquidation threshold of this user, value is in decimal, e.g.0.81, this value should be larger than ltv
        | value = (token_amount0 * LT0 + token_amount1 * LT1 + ...) / (token_amount0 + token_amount1)

        :param collaterals: collateral values of each token, note: dict value(Decimal) is token value in usd
        :type collaterals: Dict[SupplyKey, Decimal]
        :param risk_parameters: token risk parameters of this chain
        :type risk_parameters: DataFrame
        :return: total liquidation threshold
        :rtype: Decimal
        """
        sum_amount = DECIMAL_0
        rate = DECIMAL_0
        for t, s in collaterals.items():
            sum_amount += s
            rate += s * risk_parameters.loc[t.token.name].liqThereshold

        return AaveV3CoreLib.safe_div(rate, sum_amount)

    @staticmethod
    def safe_div(a: Decimal, b: Decimal) -> Decimal:
        """
        div two decimals, but will return inf if b==0
        """
        return a / b if b != 0 else Decimal("inf")

    @staticmethod
    def safe_rounding(a: Decimal, rounding: Decimal) -> Decimal:
        """
        Round decimal number, but when value is inf or nan, will not throw an exception
        """
        return a if a == Decimal("inf") or a == Decimal("nan") else a.quantize(rounding)

    @staticmethod
    def safe_div_zero(a: Decimal, b: Decimal) -> Decimal:
        """
        div two decimals, but will return 0 if b==0, only used to calculate apy
        """
        return a / b if b != 0 else Decimal(0)

    @staticmethod
    def get_apy(amounts: Dict[ActionKey, Decimal], rate_dict: Dict[TokenInfo, Decimal]) -> Decimal:
        """
        Calculate apy of all borrows or supplies

        :param amounts: supply or borrow values, note: dict value(Decimal) is token value in usd
        :type amounts: Dict[ActionKey, Decimal]
        :param rate_dict: apy of tokens
        :type rate_dict: Dict[TokenInfo, Decimal]
        :return: total apy
        :rtype: Decimal
        """
        if len(amounts) == 0:
            return DECIMAL_0
        a = Decimal(sum([amounts[key] * AaveV3CoreLib.rate_to_apy(rate_dict[key.token]) for key, amount in amounts.items()]))
        b = Decimal(sum(amounts.values()))
        return AaveV3CoreLib.safe_div_zero(a, b)  # if total amount is 0, apy should be 0



from _decimal import Decimal
from dataclasses import dataclass
from enum import Enum
from typing import Dict, NamedTuple, Union
from typing import TypeVar

import pandas as pd

from .. import TokenInfo, UnitDecimal
from ..broker import MarketBalance, MarketStatus, BaseAction, ActionTypeEnum
from ..utils import console_text

T = TypeVar("T")
K = TypeVar("K")


class InterestRateMode(Enum):
    """
    Interest rate mode
    """

    variable = 1
    stable = 2

    def __str__(self):
        return self.name

    def __repr__(self):
        return self.name


@dataclass
class ActionKey:
    """
    Abstract key for actions(supply, borrow, etc.)

    :param token: token of this position
    :type token: TokenInfo
    """

    token: TokenInfo
    """token of this position"""

    def __str__(self):
        return self.token.name

    def __hash__(self):
        return self.token.__hash__()


@dataclass
class BorrowKey(ActionKey):
    """
    key of dict for borrow actions

    :param token: token of this position
    :type token: TokenInfo
    :param interest_rate_mode: interest rate mode
    :type interest_rate_mode: InterestRateMode
    """

    interest_rate_mode: InterestRateMode
    """interest rate mode"""

    def __str__(self):
        return f"Borrow_{self.token.name}({self.interest_rate_mode.name})"

    def __repr__(self):
        return f"Borrow_{self.token.name}({self.interest_rate_mode.name})"

    def __hash__(self):
        return hash((self.token, self.interest_rate_mode))


@dataclass
class SupplyKey(ActionKey):
    """
    key of dict for supply actions

    :param token: token of this position
    :type token: TokenInfo
    """

    def __str__(self):
        return "Supply_" + self.token.name

    def __repr__(self):
        return "Supply_" + self.token.name

    def __hash__(self):
        return self.token.__hash__()


@dataclass
class SupplyInfo:
    """
    Basic info for supply, designed to kept properties inside market

    :param base_amount: Base amount of token. Note: base amount is the amount kept in aave contract. its value is amount/liquidity_index
    :type base_amount: Decimal
    :param collateral: set this supply to collateral or not
    :type collateral: bool

    """

    base_amount: Decimal
    """Base amount of token. Note: base amount is the amount kept in aave contract. its value is amount/liquidity_index"""
    collateral: bool
    """set this supply to collateral or not"""


@dataclass
class Supply:
    """
    Supply info, designed to show supplies to user

    :param token: which token is supplied
    :type token: TokenInfo
    :param base_amount: Base amount of token. Note: base amount is the amount kept in aave contract. its value is amount/liquidity_index_at_supply_moment
    :type base_amount: Decimal
    :param collateral: set this supply to collateral or not
    :type collateral: bool
    :param amount: Actual amount at this moment. value is amount * liquidity_index_at_this_moment, unit in token amount, e.g. 1 eth
    :type amount: Decimal
    :param apy: current apy (annual interest rate)
    :type apy: Decimal
    :param value: value at this moment. unit is usd, e.g. 1400 usd
    :type value: Decimal
    """

    token: TokenInfo
    """which token is supplied"""
    base_amount: Decimal
    """Base amount of token. Note: base amount is the amount kept in aave contract. its value is amount/liquidity_index_at_supply_moment"""
    collateral: bool
    """set this supply to collateral or not"""
    amount: Decimal
    """Actual amount at this moment. value is amount * liquidity_index_at_this_moment, unit in token amount, e.g. 1 eth"""
    apy: Decimal
    """current apy (annual interest rate)"""
    value: Decimal
    """value at this moment. unit is usd, e.g. 1400 usd"""


def supply_to_dataframe(supplies: Dict[SupplyKey, Supply]) -> pd.DataFrame:
    """
    convert supply dict to a dataframe
    """
    pos_dict = {
        "token": [],
        "base_amount": [],
        "collateral": [],
        "amount": [],
        "apy": [],
        "value": [],
    }
    for k, v in supplies.items():
        pos_dict["token"].append(v.token.name)
        pos_dict["base_amount"].append(console_text.format_value(v.base_amount))
        pos_dict["collateral"].append(v.collateral)
        pos_dict["amount"].append(console_text.format_value(v.amount))
        pos_dict["apy"].append(console_text.format_value(v.apy))
        pos_dict["value"].append((console_text.format_value(v.value)))
    return pd.DataFrame(pos_dict)


@dataclass
class BorrowInfo:
    """
    Basic info for borrow, designed to kept properties inside market

    :param base_amount: Base amount of token. Note: base amount is the amount kept in aave contract. its value is amount/liquidity_index
    :type base_amount: Decimal
    """

    base_amount: Decimal
    """Base amount of token. Note: base amount is the amount kept in aave contract. its value is amount/liquidity_index"""


@dataclass
class Borrow:
    """
    borrow info, designed to show borrows to user

    :param token: which token is borrowed
    :type token: TokenInfo
    :param base_amount: Base amount of token. Note: base amount is the amount kept in aave contract. its value is amount/variable_borrow_index_at_supply_moment
    :type base_amount: Decimal
    :param interest_rate_mode: Interest rate mode
    :type interest_rate_mode: InterestRateMode
    :param amount: Actual amount at this moment. value is amount * variable_borrow_index_at_this_moment, unit in token amount, e.g. 1 eth
    :type amount: Decimal
    :param apy:  current apy (annual interest rate)
    :type apy: Decimal
    :param value: value at this moment. unit is usd, e.g. 1400 usd
    :type value: Decimal
    """

    token: TokenInfo
    """which token is borrowed"""
    base_amount: Decimal
    """Base amount of token. Note: base amount is the amount kept in aave contract. its value is amount/variable_borrow_index_at_supply_moment"""
    interest_rate_mode: InterestRateMode
    """Interest rate mode"""
    amount: Decimal
    """Actual amount at this moment. value is amount * variable_borrow_index_at_this_moment, unit in token amount, e.g. 1 eth"""
    apy: Decimal
    """current apy (annual interest rate)"""
    value: Decimal
    """value at this moment. unit is usd, e.g. 1400 usd"""


def borrow_to_dataframe(supplies: Dict[BorrowKey, Borrow]) -> pd.DataFrame:
    """
    convert borrow dict to a dataframe

    """
    pos_dict = {
        "token": [],
        "base_amount": [],
        "mode": [],
        "amount": [],
        "apy": [],
        "value": [],
    }
    for k, v in supplies.items():
        pos_dict["token"].append(v.token.name)
        pos_dict["base_amount"].append(console_text.format_value(v.base_amount))
        pos_dict["mode"].append(v.interest_rate_mode.name)
        pos_dict["amount"].append(console_text.format_value(v.amount))
        pos_dict["apy"].append(console_text.format_value(v.apy))
        pos_dict["value"].append(console_text.format_value(v.value))
    return pd.DataFrame(pos_dict)


@dataclass
class AaveBalance(MarketBalance):
    """
    Asset or position value in aave market.

    :param supplies_count: count of supplies
    :type supplies_count: int
    :param borrows_count: count of borrows
    :type borrows_count: int
    :param borrows_value: total borrow value(in usd)
    :type borrows_value: Decimal
    :param supplies_value: total supply value(in usd)
    :type supplies_value: Decimal
    :param collaterals_value: total collateral value in supplies(in usd)
    :type collaterals_value: Decimal
    :param health_factor: current health factor
    :type health_factor: Decimal
    :param current_ltv: max ltv allowed, in decimal, e.g. 0.7568
    :type current_ltv: Decimal
    :param liquidation_threshold: current liquidation threshold, in decimal, e.g.0.8
    :type liquidation_threshold: Decimal
    :param supply_apy: annual interest rate of supplies
    :type supply_apy: Decimal
    :param borrow_apy: annual interest rate of borrows
    :type borrow_apy: Decimal
    :param net_apy: total annual interest rate of all supplies/borrows
    :type net_apy: Decimal
    """

    supplies_count: int
    """count of supplies"""
    borrows_count: int
    """count of borrows"""
    borrows_value: Decimal
    """total borrow value(in usd)"""
    supplies_value: Decimal
    """total supply value(in usd)"""
    collaterals_value: Decimal
    """total collateral value in supplies(in usd)"""
    health_factor: Decimal
    """current health factor"""
    current_ltv: Decimal
    """max ltv allowed, in decimal, e.g. 0.7568"""
    liquidation_threshold: Decimal
    """current liquidation threshold, in decimal, e.g.0.8"""
    supply_apy: Decimal
    """annual interest rate of supplies"""
    borrow_apy: Decimal
    """annual interest rate of borrows"""
    net_apy: Decimal
    """total annual interest rate of all supplies/borrows"""


@dataclass
class AaveTokenStatus:
    """
    Aave pool status of one token. Usually they are got form ReserveDataUpdated event of aave pool. Consider it's the transient interest rates

    :param liquidity_rate: interest rate of supply in a second
    :type liquidity_rate: Decimal
    :param stable_borrow_rate: interest rate of stable borrow in a second
    :type stable_borrow_rate: Decimal
    :param variable_borrow_rate: interest rate of variable borrow rate in a second
    :type variable_borrow_rate: Decimal
    :param liquidity_index: Decide supply amount at this moment. consider it's the average interest rate
    :type liquidity_index: Decimal
    :param variable_borrow_index: Decide borrow amount at this moment. consider it's the average borrow interest rate
    :type variable_borrow_index: Decimal
    """

    liquidity_rate: Decimal
    """interest rate of supply in a second"""
    stable_borrow_rate: Decimal
    """interest rate of stable borrow in a second"""
    variable_borrow_rate: Decimal
    """interest rate of variable borrow rate in a second"""
    liquidity_index: Decimal
    """Decide supply amount at this moment. consider it's the average interest rate"""
    variable_borrow_index: Decimal
    """Decide borrow amount at this moment. consider it's the average borrow interest rate"""


@dataclass
class AaveMarketStatus(MarketStatus):
    """
    MarketStatus properties

    :param data: pool status
    :type data: Union[pd.Series, AaveTokenStatus]
    """

    data: Union[pd.Series, AaveTokenStatus] = None
    """pool status"""


class RiskParameter:
    """
    | Risk parameter of tokens,
    | It's a column description of dataframe
    | The csv is downloaded from https://www.config.fyi/,
    | Note: some attributes in csv file are excluded.
    """

    symbol: str
    canCollateral: bool
    LTV: float
    liqThereshold: float
    liqBonus: float
    reserveFactor: float
    canBorrow: bool
    optimalUtilization: float
    canBorrowStable: bool
    debtCeiling: float
    supplyCap: float
    borrowCap: float
    eModeLtv: float
    eModeLiquidationThereshold: float
    eModeLiquidationBonus: float
    borrowableInIsolation: float


@dataclass
class SupplyAction(BaseAction):
    """
    Describe parameters and results of supply transaction

    :param token: which token is supplied
    :type token: str
    :param amount: amount supplied
    :type amount: UnitDecimal
    :param collateral: collateral the supply or not.
    :type collateral: bool
    :param deposit_after: total supply amount of this token after supply
    :type deposit_after: UnitDecimal
    """

    token: str
    """which token is supplied"""
    amount: UnitDecimal
    """amount supplied"""
    collateral: bool
    """collateral the supply or not."""
    deposit_after: UnitDecimal
    """total supply amount of this token after supply"""

    def set_type(self):
        self.action_type = ActionTypeEnum.aave_supply


@dataclass
class WithdrawAction(BaseAction):
    """
    Describe parameters and results of withdraw transaction

    :param token: which token is supplied
    :type token: str
    :param amount: amount supplied
    :type amount: UnitDecimal
    :param deposit_after: total supply amount of this token after withdraw
    :type deposit_after: UnitDecimal
    """

    token: str
    """which token is supplied"""
    amount: UnitDecimal
    """amount supplied"""
    deposit_after: UnitDecimal
    """total supply amount of this token after withdraw"""

    def set_type(self):
        self.action_type = ActionTypeEnum.aave_withdraw


@dataclass
class BorrowAction(BaseAction):
    """
    Describe parameters and results of borrow transaction

    :param token: which token is borrowed
    :type token: str
    :param interest_rate_mode: interest rate mode
    :type interest_rate_mode: InterestRateMode
    :param amount: amount borrowed
    :type amount: UnitDecimal
    :param debt_after: total borrow amount of this token after borrow transaction
    :type debt_after: UnitDecimal
    """

    token: str
    """which token is borrowed"""
    interest_rate_mode: InterestRateMode
    """interest rate mode"""
    amount: UnitDecimal
    """amount borrowed"""
    debt_after: UnitDecimal
    """total borrow amount of this token after borrow transaction"""

    def set_type(self):
        self.action_type = ActionTypeEnum.aave_borrow


@dataclass
class RepayAction(BaseAction):
    """
    Describe parameters and results of RepayAction transaction

    :param token: which token is borrowed
    :type token: str
    :param interest_rate_mode: interest rate mode
    :type interest_rate_mode: InterestRateMode
    :param amount: amount repaid
    :type amount: UnitDecimal
    :param debt_after: total borrow amount of this token after repay transaction
    :type debt_after: UnitDecimal
    """

    token: str
    """which token is borrowed"""
    interest_rate_mode: InterestRateMode
    """interest rate mode"""
    amount: UnitDecimal
    """amount repaid"""
    debt_after: UnitDecimal
    """total borrow amount of this token after repay transaction"""

    def set_type(self):
        self.action_type = ActionTypeEnum.aave_repay


@dataclass
class LiquidationAction(BaseAction):
    """
    Describe parameters and results of a liquidation

    :param collateral_token: which collateral token is used in liquidation
    :type collateral_token: str
    :param debt_token:  which debt token is used in liquidation
    :type debt_token: str
    :param delt_to_cover: Debt amount to be liquidated, should be equal to variable_delt_liquidated+stable_delt_liquidated
    :type delt_to_cover: UnitDecimal
    :param collateral_used: Collateral amount to subtract in liquidation
    :type collateral_used: UnitDecimal
    :param variable_delt_liquidated: liquidated debt token amount in variable delt
    :type variable_delt_liquidated: UnitDecimal
    :param stable_delt_liquidated: liquidated debt token amount in stable delt
    :type stable_delt_liquidated: UnitDecimal
    :param health_factor_before: health factor before liquidation
    :type health_factor_before: UnitDecimal
    :param health_factor_after: health factor after liquidation
    :type health_factor_after: UnitDecimal
    :param collateral_after: collateral token amount after liquidation
    :type collateral_after: UnitDecimal
    :param variable_debt_after: variable debt token amount after liquidation
    :type variable_debt_after: UnitDecimal
    :param stable_delt_after: stable delt token amount after liquidation
    :type stable_delt_after: UnitDecimal
    """

    collateral_token: str
    """which collateral token is used in liquidation"""
    debt_token: str
    """which debt token is used in liquidation"""
    delt_to_cover: UnitDecimal
    """Debt amount to be liquidated, should be equal to variable_delt_liquidated+stable_delt_liquidated"""
    collateral_used: UnitDecimal
    """Collateral amount to subtract in liquidation"""
    variable_delt_liquidated: UnitDecimal
    """liquidated debt token amount in variable delt"""
    stable_delt_liquidated: UnitDecimal
    """liquidated debt token amount in stable delt"""
    health_factor_before: Decimal
    """health factor before liquidation"""
    health_factor_after: Decimal
    """health factor after liquidation"""
    collateral_after: UnitDecimal
    """collateral token amount after liquidation"""
    variable_debt_after: UnitDecimal
    """variable debt token amount after liquidation"""
    stable_delt_after: UnitDecimal
    """stable delt token amount after liquidation"""

    def set_type(self):
        self.action_type = ActionTypeEnum.aave_repay


class DictCache:
    """
    A cache class in dictionary. It uses dict for storage.
    """

    def __init__(self):
        self.empty = True
        self._value: Dict[K, T] = {}

    @property
    def value(self) -> Dict[K, T]:
        """
        Get dict in this cache instance.
        """
        return self._value

    def get(self, k: K) -> T:
        """
        Get value from cache.

        :param k: key
        :type k: K
        :return: value
        :rtype: T
        """
        return self._value[k]

    def reset(self):
        """
        Reset cache instance, will set values to empty.
        """
        self._value: Dict[K, T] = {}
        self.empty = True

    def set(self, k: K, v: T):
        """
        Set value to dict

        :param k: key
        :type k: K
        :param v: value
        :type v: T
        """
        self._value[k] = v
        self.empty = False


class AaveMarketDescription(NamedTuple):
    """
    Designed to generate json description for aave market

    :param type: market type
    :type type: str
    :param name: market name
    :type name: str
    :param supplies_count: count of supplies
    :type supplies_count: int
    :param borrows_count: count of borrows
    :type borrows_count: int
    """

    type: str
    """market type"""
    name: str
    """market name"""
    supplies_count: int
    """count of supplies"""
    borrows_count: int
    """count of borrows"""



from decimal import Decimal
from enum import Enum


class Formats:
    # follow the document here: https://python-reference.readthedocs.io/en/latest/docs/functions/format.html
    global_num_format: str = ".8g"


# constant value for number 1
DECIMAL_0 = Decimal(0)

# constant value for number 0
DECIMAL_1 = Decimal(1)


class TimeUnitEnum(Enum):
    """
    Time unit of moving average,

    * minute
    * hour
    * day
    """

    minute = 1
    hour = 60
    day = 60 * 24


class UnitDecimal(Decimal):
    """
    Decimal with unit, such a 1 eth.

    It's inherit from Decimal, but considering performance issues, calculate function is not override,
    so if you do calculate on this object, return type will be Decimal

    :param number: number to keep
    :type number: Decimal
    :param unit: unit of the number, e.g. eth
    :type unit: str
    """

    __integral = Decimal(1)

    def __new__(cls, value, unit: str = ""):
        obj = Decimal.__new__(cls, value)
        obj._unit = unit
        return obj

    def to_str(self):
        """
        Get formatted string like "12.34 eth". Decimal format is predefined by self.output_format attribute

        :return: formatted string
        :rtype: str
        """
        dec = self.quantize(DECIMAL_1) if (self == self.to_integral() and self < 1e29) else self.normalize()
        return "{:{}} {}".format(dec, Formats.global_num_format, self._unit)

    @property
    def unit(self):
        return self._unit

    @unit.setter
    def unit(self, value):
        self._unit = value


class EvaluatorEnum(Enum):
    """
    Types of Strategy Evaluation
    """

    all = 0
    annualized_returns = 1
    benchmark_returns = 2
    max_draw_down = 3
    net_value = 4
    profit = 5
    net_value_up_down_rate = 6
    eth_up_down_rate = 7
    position_fee_profit = 8
    position_fee_annualized_returns = 9
    position_market_time_rate = 10

    def __str__(self):
        return self.name

    def __repr__(self):
        return self.name


class TokenInfo:
    """
    Identity for a token, will be used as key for token dict.

    :param name: token symbol, will be set as unit of a token value, e.g. usdc
    :type name: str
    :param decimal: decimal of this token, e.g. 6
    :type decimal: int
    :param address: Address of token, for aave market, this attribute has to be filled to load data.
    :type decimal: str
    """

    name: str
    decimal: int
    address: str

    def __init__(self, name: str, decimal: int, address: str = ""):
        self.name = name.upper()
        self.decimal = decimal
        self.address = address.lower()

    def __str__(self):
        return self.name

    def __repr__(self):
        return self.name

    def __eq__(self, other):
        if isinstance(other, TokenInfo):
            return self.name == other.name
        else:
            return False

    def __hash__(self):
        return self.name.__hash__()


class DemeterError(Exception):
    def __init__(self, message):
        self.message = message


class ChainType(str, Enum):
    """
    Enum for chains
    """

    ethereum = "ethereum"
    polygon = "polygon"
    optimism = "optimism"
    arbitrum = "arbitrum"
    celo = "celo"
    bsc = "bsc"
    base = "base"
    avalanche = "avalanche"
    fantom = "fantom"
    harmony = "harmony"



import math
from decimal import Decimal, getcontext
from typing import Tuple

from .liquitidy_math import get_sqrt_ratio_at_tick

Q96 = Decimal(2**96)
SQRT_1p0001 = math.sqrt(Decimal(1.0001))
getcontext().prec = 35  # default is 28, 33 is good enough for 3000


def _from_x96(number: int) -> Decimal:
    """
    decimal divide 2 ** 96

    :param number: sqrted x96 price
    :return: sqrted price
    """
    return Decimal(number) / Q96


def _to_x96(sqrt_price: Decimal) -> int:
    """
    decimal multiple 2 ** 96

    :param sqrt_price: price in base unit
    :return: result
    """
    if not isinstance(sqrt_price, Decimal):
        sqrt_price = Decimal(sqrt_price)
    return int(sqrt_price * Q96)


def sqrt_price_x96_to_base_unit_price(
    sqrt_price_x96, token_0_decimal: int, token_1_decimal: int, is_token0_quote: bool
) -> Decimal:
    """
    convert sqrt x96 price to decimal

    :param sqrt_price_x96: sqrt x96 price
    :param token_0_decimal: decimal of token 0
    :param token_1_decimal: decimal of token 1
    :param is_token0_quote: is token 0 the quote token
    :return: price in base unit, e.g. 1234.56 eth/usdc
    """
    sqrt_price = _from_x96(sqrt_price_x96)
    pool_price = sqrt_price**2 * Decimal(10 ** (token_0_decimal - token_1_decimal))

    return Decimal(1 / pool_price) if is_token0_quote else pool_price


def base_unit_price_to_sqrt_price_x96(
    price: Decimal, token_0_decimal: int, token_1_decimal: int, is_token0_quote: bool
) -> int:
    """
    convert quote price to sqrt

    :param price: price in base/quote
    :param token_0_decimal: token0 decimal
    :param token_1_decimal: token1 decimal
    :param is_token0_quote: token 0 is quote token
    :return: sqrt price x96
    """
    # quote price->add decimal pool price->sqrt_price ->ticker

    price = 1 / price if is_token0_quote else price
    atomic_unit_price = price / Decimal(10 ** (token_0_decimal - token_1_decimal))
    sqrt_price = Decimal.sqrt(atomic_unit_price)
    return _to_x96(sqrt_price)


# can round by spacing?
def sqrt_price_x96_to_tick(sqrt_price_x96: int) -> int:
    """
    convert sqrt_priceX96 to tick

    :param sqrt_price_x96: sqrt x96 price
    :return: tick price
    """
    sqrt_price = _from_x96(sqrt_price_x96)
    return _sqrt_price_to_tick(sqrt_price)


def _sqrt_price_to_tick(sqrt_price: Decimal) -> int:
    """
    pool price to tick price

    :param sqrt_price: sqrt price
    :return: tick price
    """
    return int(math.log(sqrt_price, SQRT_1p0001))


def tick_to_sqrt_price_x96(tick: int) -> int:
    """
    convert tick data to sqrt X96 price

    :param tick: tick price
    :return: sqrt x96 price
    """
    return get_sqrt_ratio_at_tick(tick)


def tick_to_base_unit_price(tick: int, token_0_decimal: int, token_1_decimal: int, is_token0_quote: bool):
    """
    get quote price from tick price

    :param tick: tick data
    :param token_0_decimal: token0 decimal
    :param token_1_decimal: token1 decimal
    :param is_token0_quote: quote on token0
    :return: quote price
    """
    sqrt_price_x96 = get_sqrt_ratio_at_tick(tick)
    atomic_unit_price = _from_x96(sqrt_price_x96) ** 2
    pool_price = atomic_unit_price * Decimal(10 ** (token_0_decimal - token_1_decimal))
    return Decimal(1 / pool_price) if is_token0_quote else pool_price


def base_unit_price_to_tick(price: Decimal, token_0_decimal: int, token_1_decimal: int, is_token0_quote: bool) -> int:
    """
    quote price to tick price

    :param price: price in base unit, e.g. 1234.56 eth/usdc
    :param token_0_decimal: token0 decimal
    :param token_1_decimal: token1 decimal
    :param is_token0_quote: token 0 is quote token
    :return: tick price
    """

    # quote price->add decimal pool price->sqrt_price ->tick
    price = 1 / price if is_token0_quote else price
    atomic_unit_price = price / Decimal(10 ** (token_0_decimal - token_1_decimal))
    sqrt_price = Decimal.sqrt(atomic_unit_price)
    return _sqrt_price_to_tick(sqrt_price)


def from_atomic_unit(atomic_unit_amount: int, decimal: int) -> Decimal:
    """
    Convert token amount to wei

    :param atomic_unit_amount: token amount
    :param decimal: decimal of token
    :return: token amount in base unit
    """
    return Decimal(int(atomic_unit_amount)) / Decimal(10**decimal)


def get_delta_gamma(
    lower_price: float,
    upper_price: float,
    price: float,
    liquidity: int,
    decimal0: int,
    decimal1: int,
    is_token0_quote: bool,
) -> Tuple[Decimal, Decimal]:
    """
    get delta gamma

    :param lower_price: lower price
    :param upper_price: upper price
    :param price: price
    :param liquidity: liquidity
    :param decimal0: decimal 0
    :param decimal1: decimal 1
    :param is_token0_quote: check if token 0 is quote
    :return: delta and gamma
    """
    lower_price_sqrt_x96 = base_unit_price_to_sqrt_price_x96(Decimal(lower_price), decimal0, decimal1, is_token0_quote)
    upper_price_sqrt_x96 = base_unit_price_to_sqrt_price_x96(Decimal(upper_price), decimal0, decimal1, is_token0_quote)
    if lower_price_sqrt_x96 > upper_price_sqrt_x96:
        (lower_price_sqrt_x96, upper_price_sqrt_x96) = (
            upper_price_sqrt_x96,
            lower_price_sqrt_x96,
        )
    return get_delta_gamma_sqrt_x96(
        lower_price,
        lower_price_sqrt_x96,
        upper_price,
        upper_price_sqrt_x96,
        price,
        liquidity,
        decimal0,
        decimal1,
        is_token0_quote,
    )


def get_delta_gamma_sqrt_x96(
    lower_price,
    sqrtA: int,
    upper_price,
    sqrtB: int,
    price,
    liquidity: int,
    d0: int,
    d1: int,
    is_token0_quote: bool,
) -> Tuple[Decimal, Decimal]:
    """
    Get delta gamma in sqrtX96 price

    """
    """
    Delta is calculated by integrating net worth, and gamma is calculated by integrating delta,
    Therefore, the most important thing is to find the calculation formula of the net value (with tick range, price, and liquidity as parameters),
    and then derive the formula after integration

    The following comment indicates how to calculate net value.

    * a: amount
    * p: decimal price of base token


    k = 2 ** 96
    a0 = k * (10**(-d)) * Liquidity * (1/SqrtPrice - 1/upper_price_sqrtX96)
    a1= Liquidity / k / 10**d * (SqrtPrice - lower_price_sqrtX96)


    if 0 quote:
    SqrtPrice=k / (10 ** (d/2)) / (p**0.5)
    net_value = a1 * p                    price <= lower, a1 is constant
                a0 + a1 * p               lower < price < upper
                a0                        price >= upper, a0 is constant

    a0 + a1 * p = liquidity * 10 ** (0.5 * d) / 10 ** d0 * price_float ** 0.5 - \
              k / upper_sqrt * liquidity / 10 ** d0 + \
              liquidity*  price_float ** 0.5 / 10 ** d1 / 10 ** (0.5 * d) - \
              lower_sqrt / k * price_float * liquidity / 10 ** d1


    if 1 quote
    SqrtPrice = k * p**0.5 / (10 ** (d/2))

    net_value = a0 * p                         price <= lower, a0 is constant
                a0 * p + a1                    lower < price < upper
                a1                             price >= upper, a1 is constant

    a0 * p + a1 = liquidity * price_float ** 0.5 * 10 ** (0.5 * d) / 10 ** d0 - \
                  k / upper_sqrt * liquidity * price_float / 10 ** d0 + \
                  liquidity * price_float ** 0.5 / 10 ** d1 / 10 ** (0.5 * d) - \
                  lower_sqrt / k * liquidity / 10 ** d1



    a0 + p * a1 =  Liquidity / 10**(d/2) / p**(1/2) + Liquidity * p**(1.5) / 10 ** (1.5*d) -
                   Liquidity * lower_price_sqrtX96 * p / 2**96 / 10**d
    """
    k = 2**96
    d = d0 - d1
    if is_token0_quote:
        if price <= lower_price:
            delta = liquidity / 2**96 / 10**d1 * (sqrtB - sqrtA)
            gamma = 0
        elif lower_price < price < upper_price:
            m = 10 ** (0.5 * d)
            delta = liquidity * (0.5 * m / price**0.5 / 10**d0 + 0.5 / 10**d1 / m / price**0.5 - sqrtA / k / 10**d1)
            gamma = -0.25 * liquidity / price**1.5 * (m / 10**d0 + 1 / 10**d1 / m)
        else:
            delta = 0
            gamma = 0
    else:
        if price <= lower_price:
            delta = liquidity / 10**d0 * (k / sqrtA - k / sqrtB)
            gamma = 0
        elif lower_price < price < upper_price:
            m = 10 ** (0.5 * d)
            delta = liquidity * (0.5 * m / price**0.5 / 10**d0 + 0.5 / 10**d1 / m / price**0.5 - k / sqrtB / 10**d0)
            gamma = -0.25 * liquidity / price**1.5 * (m / 10**d0 + 1 / m / 10**d1)
        else:
            delta = 0
            gamma = 0

    return delta, gamma



"""
Uniswap market, this module can simulate common operations in Uniswap V3, such as add/remove liquidity, swap etc.
"""

from ._typing import (
    UniV3Pool,
    UniV3PoolStatus,
    Position,
    UniLpBalance,
    PositionInfo,
    SellAction,
    BuyAction,
    RemoveLiquidityAction,
    CollectFeeAction,
    AddLiquidityAction,
    UniswapMarketStatus,
)
from .core import V3CoreLib
from .data import LineTypeEnum, UniLPData
from .market import UniLpMarket



import warnings
from decimal import Decimal

import pandas as pd


def annualized_returns(init_value, final_value, timespan_in_day):
    """
    calculated for a period of a year's data

    :param init_value:  Net value in the beginning
    :param final_value: Net value in the end
    :param timespan_in_day: time span, unit is day
    """
    return (final_value / init_value) ** Decimal(365 / timespan_in_day) - 1


def get_benchmark_returns(
    init_value: Decimal,
    init_price: pd.Series,
    final_price: pd.Series,
    timespan_in_day: Decimal,
)->Decimal:
    """
    Annualized benchmark return rate

    algorithm: swap token balance to 1:1, and hold those get_position to the end.

    :param init_value: total net value in the beginning
    :type init_value: Decimal
    :param init_price: price in the beginning, price should contain two items, base price and quote price
    :type init_price: Series
    :param final_price: Price in the end, price should contain two items, base price and quote price
    :type final_price: Series
    :param timespan_in_day: time span, unit is day
    :type timespan_in_day: Decimal
    :return: benchmark return
    :rtype: Decimal
    """
    splited_value = init_value / len(init_price)

    amounts = []
    for price in init_price:
        amounts.append(splited_value / price)
    final_value = 0
    i = 0
    for price in final_price:
        final_value += amounts[i] * price
        i += 1
    return (final_value / init_value) ** Decimal(365 / timespan_in_day) - 1


def __devide_value_to_50_50(net_value, price):
    """
    divide base/quote token value to 50:50 by according to price

    :param net_value: Decimal or float
    :param price: quote token price
    """
    base_amount = net_value / 2
    quote_amount = (net_value - base_amount) / price
    return base_amount, quote_amount


def max_draw_down(value: pd.Series):
    """
    Get max draw down

    :param value: value to calculate
    :type value: pd.Series
    """
    warnings.warn("use max_draw_down_fast instead", DeprecationWarning)
    value.index = range(len(value.index))  # restruct index to access faster
    result = 0
    for index, row in value.iteritems():
        current_max = value[index:].apply(lambda nv: 1 - nv / row).max()
        if current_max > result:
            result = current_max
    return result


def max_draw_down_fast(value: pd.Series):
    """
    Get max draw down in a fast algorithm.

    :param value: value to calculate
    :type value:  pd.Series
    """
    max_value, idx_h, idx_l = _withdraw_with_high_low(value.to_list())
    return (value.iloc[idx_h] - value.iloc[idx_l]) / value.iloc[idx_h]


def _withdraw_with_high_low(arr: list):
    """
    from : https://blog.csdn.net/Spade_/article/details/112341428
    """

    # 传入一个数组，返回最大回撤和对应的最高点索引、最低点索引
    _dp = 0  # 使用 _dp 表示 i 点的最大回撤
    i_high = 0  # 遍历时，0 ~ i - 1 中最高的点的索引，注意是索引

    # 全局最大回撤和对应的最高点和最低点的索引，注意是索引
    g_withdraw, g_high, g_low = float("-inf"), -1, -1

    for i in range(1, len(arr)):
        if arr[i_high] < arr[i - 1]:  # 若 0 ~ i - 1 中最高的点小于当前点
            i_high = i - 1  # 0 ~ i - 1 中最高的点的索引

        _dp = arr[i_high] - arr[i]  # _dp 表示 i 点的最大回撤
        if _dp > g_withdraw:  # 找到新的最大回撤，更新三个值
            g_withdraw = _dp
            g_high = i_high
            g_low = i

    return g_withdraw, g_high, g_low



import math
from decimal import Decimal
from typing import Dict

import pandas as pd

from demeter.squeeth import VaultKey, Vault
from demeter.utils import console_text


def calc_twap_price(prices: pd.Series) -> Decimal:
    """
    Calc TWAP(time weighted average price) as uniswap oracle.

    :param prices: given price array.
    :type prices: Series
    :return: TWAP price
    :rtype: Decimal
    """
    logged = prices.apply(lambda x: math.log(x, 1.0001))
    logged_sum = logged.sum()
    power = logged_sum / len(prices)
    avg_price = math.pow(1.0001, power)

    return Decimal(avg_price)


def vault_to_dataframe(vaults: Dict[VaultKey, Vault]) -> pd.DataFrame:
    """
    convert supply dict to a dataframe
    """
    vault_dict = {
        "id": [],
        "collateral_amount": [],
        "osqth_short_amount": [],
        "uni_nft_id": [],
    }
    for k, v in vaults.items():
        vault_dict["id"].append(v.id)
        vault_dict["collateral_amount"].append(console_text.format_value(v.collateral_amount))
        vault_dict["osqth_short_amount"].append(console_text.format_value(v.osqth_short_amount))
        vault_dict["uni_nft_id"].append(str(v.uni_nft_id))
    return pd.DataFrame(vault_dict)



"""
Utils functions of demeter
"""

from .application import float_param_formatter, to_decimal
from .console_text import (
    ForColorEnum,
    BackColorEnum,
    ModeEnum,
    get_formatted,
    STYLE,
    get_formatted_predefined,
    get_formatted_from_dict,
)



from dataclasses import dataclass
from decimal import Decimal
from typing import NamedTuple

from demeter import TokenInfo, ChainType, BaseAction, UnitDecimal
from demeter.broker import MarketBalance, ActionTypeEnum
from demeter.uniswap import PositionInfo

oSQTH = TokenInfo("oSQTH", 18)
WETH = TokenInfo("weth", 18)
USDC = TokenInfo("usdc", 6)


@dataclass
class SqueethChain:
    """
    Chain config of squeeth

    :param chain: Chan type
    :type chain: ChainType
    :param controller: controller contract address
    :type controller: str
    :param eth_quote_currency_pool: eth-usdc pool in uniswap,
    :type eth_quote_currency_pool: str
    :param eth_quote_currency: quote currency, on ethereum is usdc
    :type eth_quote_currency: TokenInfo
    :param squeeth_uni_pool: eth-squeeth pool in squeeth
    :type squeeth_uni_pool: str

    """
    chain: ChainType
    controller: str
    eth_quote_currency_pool: str
    eth_quote_currency: TokenInfo
    squeeth_uni_pool: str


ETH_MAINNET = SqueethChain(
    chain=ChainType.ethereum,
    controller="0x64187ae08781b09368e6253f9e94951243a493d5",
    eth_quote_currency_pool="0x8ad599c3a0ff1de082011efddc58f1908eb6e6d8",
    eth_quote_currency=TokenInfo("USDC", 6),
    squeeth_uni_pool="0x82c427adfdf2d245ec51d8046b41c4ee87f0d29c",
)


@dataclass
class Vault:
    """
    Squeeth short position

    :param id: key of vault
    :type id: int
    :param collateral_amount: eth amount in collateral
    :type collateral_amount: Decimal
    :param osqth_short_amount: osqth debt amount
    :type osqth_short_amount: Decimal
    :param uni_nft_id: uniswap lp position in collateral
    :type uni_nft_id: PositionInfo
    """
    id: int
    collateral_amount: Decimal = Decimal(0)
    osqth_short_amount: Decimal = Decimal(0)
    uni_nft_id: PositionInfo | None = None


@dataclass
class SqueethBalance(MarketBalance):
    """
    Balance of squeeth market


    :param collateral_amount: collateral amount in eth, including eth and uniswap lp
    :type collateral_amount: Decimal
    :param collateral_value: collateral value in usd,
    :type collateral_value: Decimal
    :param osqth_long_amount: osqth amount in broker
    :type osqth_long_amount: Decimal
    :param osqth_short_amount: osqth debt
    :type osqth_short_amount: Decimal
    :param osqth_short_in_eth: debt in eth. calculated by index price
    :type osqth_short_in_eth: Decimal
    :param osqth_net_amount: long amount - short amount
    :type osqth_net_amount: Decimal
    :param collateral_ratio: collateral ratio
    :type collateral_ratio: Decimal
    :param vault_count: vault count
    :type vault_count: int
    :param delta: delta
    :type delta: Decimal
    :param gamma: gamma
    :type gamma: Decimal
    """
    collateral_amount: Decimal
    collateral_value: Decimal
    osqth_long_amount: Decimal
    osqth_short_amount: Decimal
    osqth_short_in_eth: Decimal
    osqth_net_amount: Decimal
    collateral_ratio: Decimal
    vault_count: int
    delta: Decimal
    gamma: Decimal


class VaultKey(NamedTuple):
    """
    Vault key

    :param id: vault id
    :type id: int
    """

    # I have to use a Namedtuple as vault key other than int,
    # so it will not be converted into Decimal when used as a function param

    id: int


@dataclass
class VaultAction(BaseAction):
    """
    Base action of short trade

    :param vault_id: vault id
    :type vault_id: int

    """
    vault_id: int


@dataclass
class AddVaultAction(VaultAction):
    """
    Add vault action, throws when you create a new vault.

    :param vault_count: vault count
    :type vault_count: int
    """
    vault_count: int

    def set_type(self):
        self.action_type = ActionTypeEnum.squeeth_open_vault


@dataclass
class UpdateCollateralAction(VaultAction):
    """
    Update collateral with eth

    :param collateral_amount: eth amount to collate, when withdrawing, amount is negative
    :type collateral_amount: UnitDecimal
    :param collateral_after: collateral amount after deposit/withdraw
    :type collateral_after: UnitDecimal
    :param fee: fee, current it's always zero
    :type fee: UnitDecimal

    """
    collateral_amount: UnitDecimal
    collateral_after: UnitDecimal
    fee: UnitDecimal

    def set_type(self):
        self.action_type = ActionTypeEnum.squeeth_update_collateral


@dataclass
class UpdateShortAction(VaultAction):
    """
    Update osqth debt

    :param short_amount: osqth amount to borrow, when repaying, amount is negative
    :type short_amount: UnitDecimal
    :param short_after: debt amount after
    :type short_after: UnitDecimal

    """

    short_amount: UnitDecimal
    short_after: UnitDecimal

    def set_type(self):
        self.action_type = ActionTypeEnum.squeeth_update_short


@dataclass
class DepositLpAction(VaultAction):
    """
    deposit lp action

    :param position: position to deposit
    :type position: PositionInfo
    """
    position: PositionInfo

    def set_type(self):
        self.action_type = ActionTypeEnum.squeeth_deposit_lp


@dataclass
class WithdrawLpAction(VaultAction):
    """
    withdraw lp action

    :param position: position to withdraw
    :type position: PositionInfo
    """
    position: PositionInfo

    def set_type(self):
        self.action_type = ActionTypeEnum.squeeth_withdraw_lp


@dataclass
class ReduceDebtAction(VaultAction):
    """
    Reduce debt action,

    :param position: uniswap lp position redeemed
    :type position: PositionInfo
    :param withdrawn_eth_amount: eth amount withdraw from lp
    :type withdrawn_eth_amount: UnitDecimal
    :param withdrawn_osqth_amount: osqth amount withdraw from lp
    :type withdrawn_osqth_amount: UnitDecimal
    :param burn_amount: osqth burned
    :type burn_amount: UnitDecimal
    :param excess: excess
    :type excess: UnitDecimal
    :param bounty: Reduce debt bounty to pay
    :type bounty: UnitDecimal
    :param short_amount_after: debt amount after
    :type short_amount_after: UnitDecimal
    :param collateral_after: collateral amount after
    :type collateral_after: UnitDecimal
    """
    position: PositionInfo
    withdrawn_eth_amount: UnitDecimal
    withdrawn_osqth_amount: UnitDecimal
    burn_amount: UnitDecimal
    excess: UnitDecimal
    bounty: UnitDecimal
    short_amount_after: UnitDecimal
    collateral_after: UnitDecimal

    def set_type(self):
        self.action_type = ActionTypeEnum.squeeth_reduce_debt


@dataclass
class LiquidationAction(VaultAction):
    """
    Liquidation

    :param liquidate_amount: liquidated osqth debt amount
    :type liquidate_amount: UnitDecimal
    :param short_amount_after: debt amount after
    :type short_amount_after: UnitDecimal
    :param collateral_to_pay: eth collateral paid
    :type collateral_to_pay: UnitDecimal
    :param collateral_after: eth collateral after
    :type collateral_after: UnitDecimal
    """
    liquidate_amount: UnitDecimal
    short_amount_after: UnitDecimal
    collateral_to_pay: UnitDecimal
    collateral_after: UnitDecimal

    def set_type(self):
        self.action_type = ActionTypeEnum.squeeth_liquidation



from datetime import datetime
from decimal import Decimal
from typing import Dict, Callable

import pandas as pd

from ._typing import Asset, TokenInfo, AccountStatus, MarketDict, AssetDict, BaseAction
from .market import Market
from .._typing import DemeterError, UnitDecimal
from ..utils import get_formatted_from_dict, get_formatted_predefined, STYLE, float_param_formatter


class Broker:
    """
    Broker supports different order types, checking a submitted order cash requirements against current cash,
    keeping track of cash and value for each iteration of actuator and keeping the current position on different datas.

    :param allow_negative_balance: allow cash balance can be negative value or not. Default is False
    :type allow_negative_balance: bool
    :param record_action_callback: A callback function used to notify actions(buy/sell). When new actions is taken, this function will be called, and action instance will be passed as parameter. function should be like: def callback(action:BaseAction)
    :type record_action_callback: Callable[[BaseAction], None]
    """

    def __init__(self, allow_negative_balance=False, record_action_callback: Callable[[BaseAction], None] = None):
        """
        init Broker

        """
        self.allow_negative_balance = allow_negative_balance
        self._assets: AssetDict[Asset] = AssetDict()
        self._markets: MarketDict[Market] = MarketDict()
        self._record_action_callback: Callable[[BaseAction], None] = record_action_callback

    # region properties

    @property
    def markets(self) -> MarketDict[Market]:
        """
        markets managed by this broker

        :return: a dict of markets, key type is MarketKey.
        :rtype: MarketDict[Market]
        """
        return self._markets

    @property
    def assets(self) -> AssetDict[Asset]:
        """
        All assets managed by this broker.

        :return: A dict of assets, key type is TokenInfo
        :rtype: AssetDict[Asset]
        """
        return self._assets

    # endregion

    def __str__(self):
        """
        Return description of this broker in json string.

        :return: json string
        :type: str
        """
        return '{{"assets":[{}],"markets":[{}]}}'.format(
            ",".join(f"{asset}" for asset in self._assets.values()), ",".join(f"{v}" for k, v in self.markets.items())
        )

    def add_market(self, market: Market):
        """
        | Set a new market to broker,
        | User should initialize market before set to broker(Because there are too many initial parameters)

        :param market: market to add, can be uniswap market or aave market etc.
        :type market: Market
        """
        if market.market_info in self._markets:
            raise DemeterError("market has exist")
        self._markets[market.market_info] = market
        market.broker = self
        market._record_action_callback = self._record_action_callback

    @float_param_formatter
    def add_to_balance(self, token: TokenInfo, amount: Decimal | float) -> Asset:
        """
        Add token amount to wallet of broker, if token key is not exist. will generate a new record

        :param token: which token to set
        :type token: TokenInfo
        :param amount: balance, e.g. 1.2345
        :type amount: Decimal | float
        :return: Asset instance
        :rtype: Asset
        """
        if token in self._assets:
            asset: Asset = self._assets[token]
        else:
            asset: Asset = self.__add_asset(token)
        asset.add(amount)
        return asset

    @float_param_formatter
    def set_balance(self, token: TokenInfo, amount: Decimal | float) -> Asset:
        """
        Set token amount to wallet of broker. Old value will be overwritten

        :param token: which token to set, TokenInfo(name='usdc', decimal=6)
        :type token: TokenInfo
        :param amount: amount of token, e.g. 10.00
        :type amount:  Decimal | float
        :return: Asset instance
        :rtype: Asset

        """
        asset: Asset = self.__add_asset(token)
        asset.balance = amount
        return asset

    @float_param_formatter
    def subtract_from_balance(self, token: TokenInfo, amount: Decimal | float) -> Asset:
        """
        | subtract token amount from wallet of broker.
        | Wallet balance may be negative if allow_negative_balance==True
        | if token key is not exist. will generate a new record, unless broker.allow_negative_balance==False

        :param token: which token to set
        :type token: TokenInfo
        :param amount: Decimal or float type
        :type amount:  Decimal | float
        :return: Asset instance
        :rtype: Asset
        """
        if token in self._assets:
            asset: Asset = self._assets[token]
            asset.sub(amount, allow_negative_balance=self.allow_negative_balance)
        else:
            if self.allow_negative_balance:
                asset: Asset = self.__add_asset(token)
                asset.balance = 0 - amount
            else:
                raise DemeterError(f"{token.name} doesn't exist in assets dict")
        return asset

    def __add_asset(self, token: TokenInfo) -> Asset:
        self._assets[token] = Asset(token, 0)
        return self._assets[token]

    def get_token_balance(self, token: TokenInfo) -> Decimal:
        """
        Get balance of token, if token has not set, will raise an error

        :param token: which token to find.
        :type token: TokenInfo
        :return: balance of this token, e.g. 1.2345
        :rtype: Decimal
        """
        if token in self.assets:
            return self._assets[token].balance
        else:
            raise DemeterError(f"{token.name} doesn't exist in assets dict")

    def get_token_balance_with_unit(self, token: TokenInfo) -> UnitDecimal:
        """
        Get balance of token with unit

        :param token: which token to find.
        :type token: TokenInfo
        :return: balance of this token, e.g. 1.2345 eth
        :rtype: UnitDecimal
        """
        return UnitDecimal(self.get_token_balance(token), token.name)

    def get_account_status(self, prices: pd.Series | Dict[str, Decimal], timestamp=datetime | None) -> AccountStatus:
        """
        Get account status, including net value, cash balance and balance in all markets

        :param prices: current price, e.g. ('eth', Decimal('1610.553895752868641174609110')) ('usdc', 1)
        :type prices: pd.Series | Dict[str, Decimal]
        :param timestamp: current timestamp
        :type timestamp: datetime
        :return: balances
        :rtype: AccountStatus

        """
        account_status = AccountStatus(timestamp=timestamp)
        for k, v in self.markets.items():
            account_status.market_status[k] = v.get_market_balance(prices)
        account_status.market_status.set_default_key(self.markets.get_default_key())
        for k, v in self.assets.items():
            account_status.asset_balances[k] = v.balance
        asset_sum = sum([v * prices[k.name] for k, v in account_status.asset_balances.items()])
        market_sum = sum([v.net_value for v in account_status.market_status.values()])
        print('asset_sum', asset_sum, 'market_sum', market_sum)
        account_status.net_value = asset_sum + market_sum
        return account_status

    def formatted_str(self):
        """
        Get formatted broker description string to print in console

        :return: str info of broker
        :rtype: str
        """

        str_to_print = get_formatted_predefined("Token balance in broker", STYLE["header2"]) + "\n"
        balances = {}
        for asset in self._assets.values():
            balances[asset.name] = asset.balance
        str_to_print += get_formatted_from_dict(balances) + "\n"
        str_to_print += get_formatted_predefined("Position value in markets", STYLE["header2"]) + "\n"
        for market in self._markets.values():
            str_to_print += market.formatted_str() + "\n"
        return str_to_print



import decimal
from decimal import Decimal
from typing import Any

import pandas as pd

from demeter.utils import console_text


def round_decimal(num: Any, exponent: int) -> Decimal:
    """
    | Adjusting the number to a specific number of digits, eg:
    | self.assertEqual(Decimal("123500000"), round_decimal("123456789", 5))
    | self.assertEqual(Decimal("123000000"), round_decimal("123456789", 6))
    | self.assertEqual(Decimal("120"), round_decimal("123", 1))
    | self.assertEqual(Decimal("123"), round_decimal("123", 0))
    | self.assertEqual(Decimal("1.2"), round_decimal("1.23456789", -1))
    | self.assertEqual(Decimal("1.2346"), round_decimal("1.23456789", -4))


    :param num: number in any type, such as int/float/Decimal/str
    :type num: Any
    :param exponent: specific number of digits,
    :type exponent: int
    """
    if not isinstance(num, Decimal):
        num = Decimal(num)
    val = num.quantize(Decimal(f"1e{exponent}"), rounding=decimal.ROUND_HALF_UP)
    if exponent > 0:
        val = val.quantize(Decimal(0))
    return val


def position_to_df(positions) -> pd.DataFrame:
    pos_dict = {
        "instrument_name": [],
        "expiry_time": [],
        "strike_price": [],
        "type": [],
        "amount": [],
        "avg_buy_price": [],
        "buy_amount": [],
        "avg_sell_price": [],
        "sell_amount": [],
    }
    for k, v in positions.items():
        pos_dict["instrument_name"].append(console_text.format_value(v.instrument_name))
        pos_dict["expiry_time"].append(console_text.format_value(v.expiry_time))
        pos_dict["strike_price"].append(console_text.format_value(v.strike_price))
        pos_dict["type"].append(console_text.format_value(v.type))
        pos_dict["amount"].append(console_text.format_value(v.amount))
        pos_dict["avg_buy_price"].append(console_text.format_value(v.avg_buy_price))
        pos_dict["buy_amount"].append(console_text.format_value(v.buy_amount))
        pos_dict["avg_sell_price"].append(console_text.format_value(v.avg_sell_price))
        pos_dict["sell_amount"].append(console_text.format_value(v.sell_amount))

    return pd.DataFrame(pos_dict)



import os
from _decimal import Decimal
from typing import Dict

import pandas as pd

from demeter import DemeterError
from demeter.aave._typing import RiskParameter

MIN_TOKEN_VALUE = (1e-18) - (1e-27)


def sub_base_amount(old_v, value):
    """
    | Subtract base amount.
    | Usually, max token decimal is 18. so if token value is below 1e-18, we consider it's 0
    | It's necessary because number here is kept in decimal instead of int256. so there will be calculation errors.
    """

    new_v = old_v - value
    if new_v < MIN_TOKEN_VALUE:
        return 0
    else:
        return new_v


def load_risk_parameter(token_setting_path:str) -> pd.DataFrame | Dict[str, RiskParameter]:
    """
    | Load risk parameter, and convert float to decimal.
    | Files are downloaded from: https://www.config.fyi, please choose the right chain

    :param token_setting_path: risk parameter file path
    :type token_setting_path: str
    :return: risk parameter dataframe
    :rtype: DataFrame | Dict[str, RiskParameter]
    """
    path = os.path.join(token_setting_path)
    if not os.path.exists(path):
        raise DemeterError(f"risk parameter file {path} not exist, please download csv from https://www.config.fyi/")
    rp = pd.read_csv(path, sep=";")
    rp = rp[
        [
            "symbol",
            "canCollateral",
            "LTV",
            "liqThereshold",
            "liqBonus",
            "reserveFactor",
            "canBorrow",
            "optimalUtilization",
            "canBorrowStable",
            "debtCeiling",
            "supplyCap",
            "borrowCap",
            "eModeLtv",
            "eModeLiquidationThereshold",
            "eModeLiquidationBonus",
            "borrowableInIsolation",
        ]
    ]

    rp["LTV"] = rp["LTV"].str.rstrip("%").apply(lambda x: Decimal(x)) / 100
    rp["liqBonus"] = rp["liqBonus"].str.rstrip("%").apply(lambda x: Decimal(x)) / 100
    rp["liqThereshold"] = rp["liqThereshold"].str.rstrip("%").apply(lambda x: Decimal(x)) / 100
    rp = rp.set_index("symbol")
    return rp



from pathlib import Path

from setuptools import setup, find_packages

this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()

setup(
    name="zelos-demeter",
    version="0.5.1",
    packages=find_packages(exclude=["tests", "tests.*", "samples", "samples.*"]),
    url="https://zelos-demeter.readthedocs.io",
    license="MIT",
    author="zelos research",
    author_email="zelos@antalpha.com",
    description="better DEFI backtesting tool",
    long_description=long_description,
    long_description_content_type="text/markdown",
    python_requires=">=3.10",
    install_requires=[
        "numpy==1.26.4",
        "pandas==2.2.0",
        "python-dateutil==2.9.0.post0",
        "pytz==2024.1",
        "six==1.16.0",
        "db-dtypes==1.2.0",
        "tqdm==4.66.2",
        "orjson==3.9.15",
    ],
)

# rm -rf ./demeter.egg-info/ && python setup.py sdist upload -r private
# python setup.py sdist
# twine upload -r pypi dist/*



import logging
import os
import pickle
import time
from dataclasses import dataclass, field
from datetime import datetime
from decimal import Decimal
from typing import List, Dict, Union

import orjson
import pandas as pd
from pandas import Timestamp
from tqdm import tqdm  # process bar

from .evaluating_indicator import Evaluator
from .. import Broker, Asset
from .._typing import DemeterError, EvaluatorEnum, UnitDecimal
from ..broker import BaseAction, AccountStatus, MarketInfo, MarketDict, MarketStatus, RowData
from ..strategy import Strategy
from ..uniswap import UniLpMarket, PositionInfo
from ..utils import get_formatted_predefined, STYLE, to_decimal
from ..utils import console_text


class Actuator(object):
    """
    Core component of a back test. Manage the resources in a test, including broker/strategy/data/indicator,

    :param allow_negative_balance: Allow cash balance of broker can be negative value or not. Default is False
    :type allow_negative_balance: bool
    """

    def __init__(self, allow_negative_balance=False):
        """
        init Actuator
        """
        # all the actions during the test(buy/sell/add liquidity)
        self._action_list: List[BaseAction] = []
        self._currents = Currents()
        # broker status in every bar, use array for performance
        self._account_status_list: List[AccountStatus] = []
        self._account_status_df: pd.DataFrame | None = None

        # broker
        self._broker: Broker = Broker(allow_negative_balance, self._record_action_list)
        # strategy
        self._strategy: Strategy = Strategy()
        self._token_prices: pd.DataFrame | None = None
        # path of source data, which is saved by downloader
        # evaluating indicator calculator
        self._evaluator: Evaluator | None = None
        self._enabled_evaluator: [] = []
        # logging
        logging.basicConfig(level=logging.INFO, format="%(asctime)s - %(levelname)s - %(message)s")
        self.logger = logging.getLogger(__name__)
        # internal var
        self.__backtest_finished = False

    def _record_action_list(self, action: BaseAction):
        """
        record action list

        :param action: action
        :type action: BaseAction
        """
        action.timestamp = self._currents.timestamp
        action.set_type()
        self._action_list.append(action)
        self._currents.actions.append(action)

    # region property
    @property
    def account_status(self) -> List[AccountStatus]:
        """
        | Get account status list.
        | Account status includes balances, net values and positions.
        | Each element in this list stands for one minute.
        | It is recommended to use get_account_status_dataframe.
        """
        return self._account_status_list

    @property
    def token_prices(self) -> pd.DataFrame:
        """
        Price of all tokens. Row(index) is minutely timestamp, column is token. e.g.

        +-------------------+---------+-------+
        |                   |WETH     | USDC  |
        +===================+=========+=======+
        |2023-08-13 00:00:00| 1848.12 | 1     |
        +-------------------+---------+-------+
        |2023-08-13 00:01:00| 1848.12 | 1     |
        +-------------------+---------+-------+

        :return: data from with prices of all token
        :rtype: DataFrame
        """
        return self._token_prices

    @property
    def final_status(self) -> AccountStatus:
        """
        | Get last account status of back test.
        | Note: If back test has not run, an error will be raised.

        :return: Final state of broker
        :rtype: AccountStatus
        """
        if self.__backtest_finished:
            return self._account_status_list[len(self._account_status_list) - 1]
        else:
            raise DemeterError("please run strategy first")

    def reset(self):
        """
        Reset actuator by re-initiate all the status variables
        """
        self._evaluator: Evaluator | None = None
        self._enabled_evaluator: [] = []

        self._action_list = []
        self._currents = Currents()
        self._account_status_list = []
        self.__backtest_finished = False

        self._account_status_df: pd.DataFrame | None = None

    @property
    def actions(self) -> List[BaseAction]:
        """
        A list of actions(buy/sell/add liquidity) happened during back test

        :return: A list of actions
        :rtype: List[BaseAction]
        """
        return self._action_list

    @property
    def evaluating_indicator(self) -> Dict[EvaluatorEnum, UnitDecimal]:
        """
        Get evaluating indicator result

        :return: Result of each evaluating indicator in dict
        :rtype: Dict[EvaluatorEnum, UnitDecimal]
        """
        return self._evaluator.result if self._evaluator is not None else None

    @property
    def broker(self) -> Broker:
        """
        | Get broker instance.
        | Brokers are managers of assets. It manages cash and markets(the place to invest assets)

        """
        return self._broker

    @property
    def strategy(self) -> Strategy:
        """
        Get strategy instance

        :return: strategy instance
        :rtype: Strategy
        """
        return self._strategy

    @strategy.setter
    def strategy(self, value):
        """
        Set strategy instance to actuator
        :param value: strategy instance
        :type value: Strategy
        """
        if isinstance(value, Strategy):
            self._strategy = value
        else:
            raise ValueError()

    # endregion

    def get_account_status_dataframe(self) -> pd.DataFrame:
        """
        | Get account status in dataframe. it contains account balance/position change of every minute.
        | Row(datetimeindex) is per minute.
        | Column is net value/positions.

        :return: account status
        :rtype: DataFrame
        """
        return AccountStatus.to_dataframe(self._account_status_list)

    def set_assets(self, assets: List[Asset]):
        """
        Set initial token balance. That's cash held by the broker.

        :param assets: assets to set.
        :type assets: [Asset]
        """
        for asset in assets:
            self._broker.set_balance(asset.token_info, asset.balance)

    def set_price(self, prices: Union[pd.DataFrame, pd.Series]):
        """
        | Set price to actuator. param price can be dataframe(price of several tokens) or series(price of one token).
        | It's index time range should be larger than or equal to data.
        | And column name should be the same to token.name in upper case. e.g.

        +-------------------+---------+-------+
        |                   |WETH     | USDC  |
        +===================+=========+=======+
        |2023-08-13 00:00:00| 1848.12 | 1     |
        +-------------------+---------+-------+
        |2023-08-13 00:01:00| 1848.12 | 1     |
        +-------------------+---------+-------+

        :param prices: dataframe or series contains prices
        :type prices: Union[pd.DataFrame, pd.Series]
        """
        prices = prices.map(lambda y: to_decimal(y))
        if isinstance(prices, pd.DataFrame):
            if self._token_prices is None:
                self._token_prices = prices
            else:
                self._token_prices = pd.concat([self._token_prices, prices])
        else:
            if self._token_prices is None:
                self._token_prices = pd.DataFrame(data=prices, index=prices.index)
            else:
                self._token_prices[prices.name] = prices

    def notify(self, strategy: Strategy, actions: List[BaseAction]):
        """
        Call strategy.notify() when new action happens.

        :param strategy: Strategy instance
        :type strategy: Strategy
        :param actions: action list
        :type actions: List[BaseAction]
        """
        if len(actions) < 1:
            return
        # last_time = datetime(1970, 1, 1)
        for action in actions:
            # if last_time != action.timestamp:
            #     print(f"\033[7;34m{action.timestamp} \033[0m")
            #     last_time = action.timestamp
            strategy.notify(action)
            pass

    def _check_backtest(self):
        """
        check backtest result, including index of data, prices
        """
        # ensure a market exist
        if len(self._broker.markets) < 1:
            raise DemeterError("No market assigned")
        # ensure all token has price list.
        if self._token_prices is None:
            # if price is not set and market is uni_lp_market, get price from market automatically
            for market in self.broker.markets.values():
                if isinstance(market, UniLpMarket):
                    self.set_price(market.get_price_from_data())
            if self._token_prices is None:
                raise DemeterError("token prices is not set")
        # for token in self._broker.assets.keys():  # dict_keys([TokenInfo(name='usdc', decimal=6), TokenInfo(name='eth', decimal=18)])
        #     if token.name not in self._token_prices:
        #         raise DemeterError(f"Price of {token.name} has not set yet")

        data_length = []  # [1440]
        for market in self._broker.markets.values():
            data_length.append(len(market.data.index))
            market.check_market()  # check each market, including assets
        # ensure data length same
        # if List.count(data_length, data_length[0]) != len(data_length):
        #     raise DemeterError("data length among markets are not same")
        default_market_data = self._broker.markets.default.data
        if (
            self._token_prices.head(1).index[0] > default_market_data.head(1).index.get_level_values(0).unique()[0]
            or self._token_prices.tail(1).index[0] < default_market_data.tail(1).index.get_level_values(0).unique()[0]
        ):
            raise DemeterError("Time range of price doesn't cover market data")

        # ensure data interval same
        # length = data_length[0]
        # data_interval = []
        # if length > 1:
        #     for market in self._broker.markets.values():
        #         data_interval.append(market.data.index[1] - market.data.index[0])
        #     if List.count(data_interval, data_interval[0]) != len(data_interval):
        #         raise DemeterError("data interval among markets are not same")
        #     price_interval = self._token_prices.index[1] - self._token_prices.index[0]
        #     if price_interval != data_interval[0]:
        #         raise DemeterError("price list interval and data interval are not same")

    def __get_row_data(self, timestamp, row_id, current_price) -> RowData:
        row_data = RowData(timestamp.to_pydatetime(), row_id, current_price)
        for market_info, market in self.broker.markets.items():
            row_data.market_status[market_info] = market.market_status.data
        row_data.market_status.set_default_key(self.broker.markets.get_default_key())
        return row_data

    def __set_market_timestamp(self, timestamp: Timestamp, update: bool = False):
        """
        set markets row data
        :param timestamp:
        :param update: enable or disable has_update flag in markets, if set to false, will always update, if set to true, just update when necessary
        :return:
        """

        for market_key in self.broker.markets.keys():
            if (not update) or (update and self._broker.markets[market_key].has_update):
                ms = MarketStatus(timestamp, None)
                self._broker.markets[market_key].set_market_status(ms, self._token_prices.loc[timestamp])

    def get_test_range(self):
        longest_data = max(map(lambda m: len(m.data.index.get_level_values(0).unique()), self._broker.markets.values()))
        largest_market = list(
            filter(
                lambda m: len(m.data.index.get_level_values(0).unique()) == longest_data, self._broker.markets.values()
            )
        )[0]
        # start = largest_market.data.head(1).index.get_level_values(0).unique()
        # end = largest_market.data.tail(1).index.get_level_values(0).unique()

        return largest_market.data.index.get_level_values(0).unique()

    def run(self, evaluator: List[EvaluatorEnum] | None = None, output: bool = True):
        """
        Start back test, the whole process including:

        * reset actuator
        * initialize strategy (set object to strategy, then run strategy.initialize())
        * process each row in data
            * prepare data in this iteration
            * run trigger
            * run strategy.on_bar()
            * update market, e.g. calculate fee earned
            * run strategy.after_bar()
            * get latest account status
            * notify actions
        * run evaluator indicator
        * run strategy.finalize()
        * output result if required

        :param evaluator: enable evaluating indicator.
        :type evaluator: List[EvaluatorEnum]
        :param output: enable output.
        :type output: bool
        """
        evaluator = evaluator if evaluator is not None else []
        run_begin_time = time.time()  # 1681718968.267463
        self.reset()

        self._enabled_evaluator = evaluator
        self._check_backtest()
        index_array: pd.DatetimeIndex = (
            self.get_test_range()
        )  # list(self._broker.markets.values())[0].data.index.get_level_values(0).unique()
        self.logger.info("init strategy...")

        # set initial status for strategy, so user can run some calculation in initial function.
        self.__set_market_timestamp(index_array[0], False)
        self._currents.timestamp = index_array[0].to_pydatetime()
        # keep initial balance for evaluating
        init_account_status = self._broker.get_account_status(
            self._token_prices.head(1).iloc[0], index_array[0].to_pydatetime()
        )
        self.init_strategy()
        row_id = 0
        data_length = len(index_array)
        self.logger.info("start main loop...")
        with tqdm(total=data_length, ncols=150) as pbar:
            for timestamp_index in index_array:
                current_price = self._token_prices.loc[timestamp_index]
                # prepare data of a row

                self.__set_market_timestamp(timestamp_index, False)
                # execute strategy, and some calculate
                self._currents.timestamp = timestamp_index.to_pydatetime()
                row_data = self.__get_row_data(timestamp_index, row_id, current_price)
                if self._strategy.triggers:
                    for trigger in self._strategy.triggers:
                        if trigger.when(row_data):
                            trigger.do(row_data)

                for market in self.broker.markets.values():
                    if market.is_open and market.open is not None:
                        market.open(row_data)

                self._strategy.on_bar(row_data)

                # important, take uniswap market for example,
                # if liquidity has changed in the head of this minute, this will add the new liquidity to total_liquidity in current minute.
                self.__set_market_timestamp(timestamp_index, True)

                # update broker status, e.g. re-calculate fee
                # and read the latest status from broker
                for market in self._broker.markets.values():
                    market.update()

                row_data = self.__get_row_data(timestamp_index, row_id, current_price)
                self._strategy.after_bar(row_data)

                self._account_status_list.append(
                    self._broker.get_account_status(current_price, timestamp_index.to_pydatetime())
                )
                # notify actions in current loop
                self.notify(self.strategy, self._currents.actions)
                self._currents.actions = []
                # move forward for process bar and index
                pbar.update()
                row_id += 1

        self.logger.info("main loop finished")
        self._account_status_df: pd.DataFrame = self.get_account_status_dataframe()

        if len(self._enabled_evaluator) > 0:
            self.logger.info("Start calculate evaluating indicator...")
            self._evaluator = Evaluator(
                init_account_status, self._account_status_df, self._token_prices, self._broker.markets
            )
            self._evaluator.run(self._enabled_evaluator)
            self.logger.info("Evaluating indicator has finished it's job.")
        self._strategy.finalize()
        self.__backtest_finished = True
        if output:
            self.output()

        self.logger.info(f"Backtesting finished, execute time {time.time() - run_begin_time}s")

    def output(self):
        """
        Print backtest result to console, and it will print the following content

        1. Account status (including balances and positions) at the end of the back test.
        2. Balance change during back test.
        3. If evaluating indicator is enabled, will print evaluating of strategy.
        """
        if not self.__backtest_finished:
            raise DemeterError("Please run strategy first")
        self.logger.info(f"Print actuator summary")
        print(get_formatted_predefined("Final account status", STYLE["header1"]))
        print(self.broker.formatted_str())
        print(get_formatted_predefined("Account balance history", STYLE["header1"]))
        console_text.print_dataframe_with_precision(self._account_status_df)
        if len(self._enabled_evaluator) > 0:
            print("Evaluating indicator")
            print(self._evaluator)

    def save_result(self, path: str, account=True, actions=True) -> List[str]:
        """
        Save backtesting result

        :param path: path to save
        :type path: str
        :param account: Save account status or not, If true, it will save a csv with all balance changes.
        :type account: bool
        :param actions: Save actions or not. If true, it will save all actions in json(friendly to human) and pkl(can be loaded by python)
        :type actions: bool
        :return: A list of saved file path
        :rtype: List[str]
        """
        if not self.__backtest_finished:
            raise DemeterError("Please run strategy first")
        file_name_head = "backtest-" + datetime.now().strftime("%Y%m%d-%H%M%S")
        if not os.path.exists(path):
            os.mkdir(path)
        file_list = []
        if account:
            file_name = os.path.join(path, file_name_head + ".account.csv")
            self._account_status_df.to_csv(file_name)
            file_list.append(file_name)
        if actions:
            # save pkl file to load again
            pkl_name = os.path.join(path, file_name_head + ".action.pkl")
            with open(pkl_name, "wb") as outfile1:
                pickle.dump(self._action_list, outfile1)
            # save json to read
            actions_json_str = orjson.dumps(self._action_list, option=orjson.OPT_INDENT_2, default=_json_default)
            json_name = os.path.join(path, file_name_head + ".action.json")
            with open(json_name, "wb") as outfile:
                outfile.write(actions_json_str)

            file_list.append(json_name)
            file_list.append(pkl_name)

        self.logger.info(f"files have saved to {','.join(file_list)}")
        return file_list

    def init_strategy(self):
        """
        initialize strategy, set property to strategy. and run strategy.initialize()
        """
        if not isinstance(self._strategy, Strategy):
            raise DemeterError("strategy must be inherit from Strategy")
        self._strategy.broker = self._broker
        self._strategy.markets = self._broker.markets
        market_datas = MarketDict()
        for k, v in self.broker.markets.items():
            market_datas[k] = v.data
        market_datas.set_default_key(self.broker.markets.get_default_key())
        self._strategy.data = market_datas
        self._strategy.prices = self._token_prices
        self._strategy.account_status = self._account_status_list
        self._strategy.actions = self._action_list
        self._strategy.assets = self.broker.assets
        self._strategy.get_account_status_dataframe = self.get_account_status_dataframe
        for k, v in self.broker.markets.items():
            setattr(self._strategy, k.name, v)
        for k, v in self.broker.assets.items():
            setattr(self._strategy, k.name, v)
        self._strategy.initialize()

    def __str__(self):
        return '{{"Account status":{}, "action_count":{}, "timestamp":"{}", "strategy":"{}", "price_df_rows":{}, "price_assets":{} }}'.format(
            str(self.broker),
            len(self._action_list),
            self._currents.timestamp,
            type(self._strategy).__name__,
            len(self._token_prices.index) if self._token_prices is not None else 0,
            (
                "[" + ",".join(f'"{x}"' for x in self._token_prices.columns) + "]"
                if self._token_prices is not None
                else str([])
            ),
        )


@dataclass
class Currents:
    """
    Values in current timestamp.

    """

    actions: List[BaseAction] = field(default_factory=list)
    """Actions in this iteration"""
    timestamp: datetime = None
    """Current timestamp"""


def _json_default(obj):
    """
    format json data

    :param obj:
    :return:
    """
    if isinstance(obj, UnitDecimal):
        return obj.to_str()
    elif isinstance(obj, Decimal):
        return str(obj)
    elif isinstance(obj, MarketInfo):
        return f"{obj.name}({obj.type.name})"
    elif isinstance(obj, PositionInfo):
        return {"lower_tick": obj.lower_tick, "upper_tick": obj.upper_tick}
    else:
        raise TypeError
